import functools
import math
import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Sequence, Dict, Union, ClassVar, Optional, Set
import dataclasses

from ..misc.encodings import RadialBasis
from ...utils.activations import activation_from_str


try:
    import e3nn_jax as e3nn

    E3NN_AVAILABLE = True
    E3NN_EXCEPTION = None
    Irreps = e3nn.Irreps
    Irrep = e3nn.Irrep
except Exception as e:
    E3NN_AVAILABLE = False
    E3NN_EXCEPTION = e
    e3nn = None

    class Irreps(tuple):
        pass

    class Irrep(tuple):
        pass


class MACE(nn.Module):
    """MACE equivariant message passing neural network.

    adapted from MACE-jax github repo by M. Geiger and I. Batatia
    
    T. Plé reordered some operations and changed defaults to match the recent mace-torch version 
    -> compatibility with pretrained torch models requires some work on the parameters:
        - normalization of activation functions in e3nn differ between jax and pytorch => need rescaling
        - multiplicity ordering and signs of U matrices in SymmetricContraction differ => need to reorder and flip signs in the weight tensors
        - we use a maximum Z instead of a list of species => need to adapt species-dependent parameters
    
    References:
        - I. Batatia et al. "MACE: Higher order equivariant message passing neural networks for fast and accurate force fields." Advances in Neural Information Processing Systems 35 (2022): 11423-11436.
        https://doi.org/10.48550/arXiv.2206.07697
        - I. Batatia et al. "The design space of e(3)-equivariant atom-centered interatomic potentials." arXiv preprint arXiv:2205.06643 (2022).
        https://doi.org/10.48550/arXiv.2205.06643

    """
    _graphs_properties: Dict
    output_irreps: Union[Irreps, str] = "1x0e"
    """The output irreps of the model."""
    hidden_irreps: Union[Irreps, str] = "128x0e + 128x1o"
    """The hidden irreps of the model."""
    readout_mlp_irreps: Union[Irreps, str] = "16x0e"
    """The hidden irreps of the readout MLP."""
    graph_key: str = "graph"
    """The key in the input dictionary that corresponds to the molecular graph to use."""
    output_key: Optional[str] = None
    """The key of the embedding in the output dictionary."""
    avg_num_neighbors: float = 1.0
    """The expected average number of neighbors."""
    ninteractions: int = 2
    """The number of interaction layers."""
    num_features: Optional[int] = None
    """The number of features per node. default gcd of hidden_irreps multiplicities"""
    radial_basis: dict = dataclasses.field(
        default_factory=lambda: {"basis": "bessel", "dim": 8, "trainable": False}
    )
    """The dictionary of parameters for radial basis functions. See `fennol.models.misc.encodings.RadialBasis`."""
    lmax: int = 1
    """The maximum angular momentum to consider."""
    correlation: int = 3
    """The correlation order at each layer."""
    activation: str = "silu"
    """The activation function to use."""
    symmetric_tensor_product_basis: bool = False
    """Whether to use the symmetric tensor product basis."""
    interaction_irreps: Union[Irreps, str] = "o3_restricted"
    skip_connection_first_layer: bool = True
    radial_network_hidden: Sequence[int] = dataclasses.field(
        default_factory=lambda: [64, 64, 64]
    )
    scalar_output: bool = False
    zmax: int = 86
    """The maximum atomic number to consider."""
    convolution_mode: int = 1

    FID: ClassVar[str] = "MACE"

    @nn.compact
    def __call__(self, inputs):
        if not E3NN_AVAILABLE:
            raise E3NN_EXCEPTION

        species_indices = inputs["species"]
        graph = inputs[self.graph_key]
        distances = graph["distances"]
        vec = e3nn.IrrepsArray("1o", graph["vec"])
        switch = graph["switch"]
        edge_src = graph["edge_src"]
        edge_dst = graph["edge_dst"]

        output_irreps = e3nn.Irreps(self.output_irreps)
        hidden_irreps = e3nn.Irreps(self.hidden_irreps)
        readout_mlp_irreps = e3nn.Irreps(self.readout_mlp_irreps)

        # extract or set num_features
        if self.num_features is None:
            num_features = functools.reduce(math.gcd, (mul for mul, _ in hidden_irreps))
            hidden_irreps = e3nn.Irreps(
                [(mul // num_features, ir) for mul, ir in hidden_irreps]
            )
        else:
            num_features = self.num_features

        # get interaction irreps
        if self.interaction_irreps == "o3_restricted":
            interaction_irreps = e3nn.Irreps.spherical_harmonics(self.lmax)
        elif self.interaction_irreps == "o3_full":
            interaction_irreps = e3nn.Irreps(e3nn.Irrep.iterator(self.lmax))
        else:
            interaction_irreps = e3nn.Irreps(self.interaction_irreps)
        convol_irreps = num_features * interaction_irreps

        # convert species to internal indices
        # maxidx = max(PERIODIC_TABLE_REV_IDX.values())
        # conv_tensor = [0] * (maxidx + 2)
        # if isinstance(self.species_order, str):
        #     species_order = [el.strip() for el in self.species_order.split(",")]
        # else:
        #     species_order = [el for el in self.species_order]
        # for i, s in enumerate(species_order):
        #     conv_tensor[PERIODIC_TABLE_REV_IDX[s]] = i
        # species_indices = jnp.asarray(conv_tensor, dtype=jnp.int32)[species]
        num_species = self.zmax + 2

        # species encoding
        encoding_irreps: e3nn.Irreps = (
            (num_features * hidden_irreps).filter("0e").regroup()
        )
        species_encoding = self.param(
            "species_encoding",
            lambda key, shape: jax.nn.standardize(
                jax.random.normal(key, shape, dtype=jnp.float32)
            ),
            (num_species, encoding_irreps.dim),
        )[species_indices]
        # convert to IrrepsArray
        node_feats = e3nn.IrrepsArray(encoding_irreps, species_encoding)

        # radial embedding
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_embedding = (
            RadialBasis(
                **{
                    **self.radial_basis,
                    "end": cutoff,
                    "name": f"RadialBasis",
                }
            )(distances)
            * switch[:, None]
        )

        # spherical harmonics
        assert self.convolution_mode in [0,1,2], "convolution_mode must be 0, 1 or 2"
        if self.convolution_mode == 0:
            Yij = e3nn.spherical_harmonics(range(0, self.lmax + 1), vec, True)
        elif self.convolution_mode == 1:
            Yij = e3nn.spherical_harmonics(range(1, self.lmax + 1), vec, True)

        outputs = []
        node_feats_all = []
        for layer in range(self.ninteractions):
            first = layer == 0
            last = layer == self.ninteractions - 1

            layer_irreps = num_features * (
                hidden_irreps if not last else hidden_irreps.filter(output_irreps)
            )

            # Linear skip connection
            sc = None
            if not first or self.skip_connection_first_layer:
                sc = e3nn.flax.Linear(
                    layer_irreps,
                    num_indexed_weights=num_species,
                    name=f"skip_tp_{layer}",
                    force_irreps_out=True,
                )(species_indices, node_feats)

            ################################################
            # Interaction block (Message passing convolution)
            node_feats = e3nn.flax.Linear(node_feats.irreps, name=f"linear_up_{layer}")(
                node_feats
            )


            messages = node_feats[edge_src]
            if self.convolution_mode == 0:
                messages = e3nn.tensor_product(
                    messages,
                    Yij,
                    filter_ir_out=convol_irreps,
                    regroup_output=True,
                )
            elif self.convolution_mode == 1:
                messages = e3nn.concatenate(
                    [
                        messages.filter(convol_irreps),
                        e3nn.tensor_product(
                            messages,
                            Yij,
                            filter_ir_out=convol_irreps,
                        ),
                        # e3nn.tensor_product_with_spherical_harmonics(
                        #     messages, vectors, self.max_ell
                        # ).filter(convol_irreps),
                    ]
                ).regroup()
            else:
                messages = e3nn.tensor_product_with_spherical_harmonics(
                    messages, vec, self.lmax
                ).filter(convol_irreps).regroup()

            # mix = FullyConnectedNet(
            #     [*self.radial_network_hidden, messages.irreps.num_irreps],
            #     activation=activation_from_str(self.activation),
            #     name=f"radial_network_{layer}",
            #     use_bias=False,
            # )(radial_embedding)
            mix = e3nn.flax.MultiLayerPerceptron(
                [*self.radial_network_hidden, messages.irreps.num_irreps],
                act=activation_from_str(self.activation),
                output_activation=False,
                name=f"radial_network_{layer}",
                gradient_normalization="element",
            )(
                radial_embedding
            )

            messages = messages * mix
            node_feats = (
                e3nn.IrrepsArray.zeros(
                    messages.irreps, node_feats.shape[:1], messages.dtype
                )
                .at[edge_dst]
                .add(messages)
            )
            # print("irreps_mid jax",node_feats.irreps)
            # jax.debug.print("node_feats={n}", n=jnp.sum(node_feats.array,axis=0)[550:570])

            node_feats = (
                e3nn.flax.Linear(convol_irreps, name=f"linear_dn_{layer}")(node_feats)
                / self.avg_num_neighbors
            )

            if first and not self.skip_connection_first_layer:
                node_feats = e3nn.flax.Linear(
                    node_feats.irreps,
                    num_indexed_weights=num_species,
                    name=f"skip_tp_{layer}",
                )(species_indices, node_feats)

            ################################################
            # Equivariant product basis block

            # symmetric contractions
            node_feats = SymmetricContraction(
                keep_irrep_out={ir for _, ir in layer_irreps},
                correlation=self.correlation,
                num_species=num_species,
                gradient_normalization="element",  # NOTE: This is to copy mace-torch
                symmetric_tensor_product_basis=self.symmetric_tensor_product_basis,
            )(
                node_feats, species_indices
            )


            node_feats = e3nn.flax.Linear(
                layer_irreps, name=f"linear_contraction_{layer}"
            )(node_feats)


            if sc is not None:
                # add skip connection
                node_feats = node_feats + sc


            ################################################
            
            # Readout block
            if last:
                num_vectors = readout_mlp_irreps.filter(drop=["0e", "0o"]).num_irreps
                layer_out = e3nn.flax.Linear(
                    (readout_mlp_irreps + e3nn.Irreps(f"{num_vectors}x0e")).simplify(),
                    name=f"hidden_linear_readout_last",
                )(node_feats)
                layer_out = e3nn.gate(
                    layer_out,
                    even_act=activation_from_str(self.activation),
                    even_gate_act=None,
                )
                layer_out = e3nn.flax.Linear(
                    output_irreps, name=f"linear_readout_last"
                )(layer_out)
            else:
                layer_out = e3nn.flax.Linear(
                    output_irreps,
                    name=f"linear_readout_{layer}",
                )(node_feats)

            if self.scalar_output:
                layer_out = layer_out.filter("0e").array

            outputs.append(layer_out)
            node_feats_all.append(node_feats.filter("0e").array)

        if self.scalar_output:
            output = jnp.stack(outputs, axis=1)
        else:
            output = e3nn.stack(outputs, axis=1)

        node_feats_all = jnp.concatenate(node_feats_all, axis=-1)

        output_key = self.output_key if self.output_key is not None else self.name
        return {
            **inputs,
            output_key: output,
            output_key + "_node_feats": node_feats_all,
        }


class SymmetricContraction(nn.Module):

    correlation: int
    keep_irrep_out: Set[Irrep]
    num_species: int
    gradient_normalization: Union[str, float]
    symmetric_tensor_product_basis: bool

    @nn.compact
    def __call__(self, input, index):
        if not E3NN_AVAILABLE:
            raise E3NN_EXCEPTION

        if self.gradient_normalization is None:
            gradient_normalization = e3nn.config("gradient_normalization")
        else:
            gradient_normalization = self.gradient_normalization
        if isinstance(gradient_normalization, str):
            gradient_normalization = {"element": 0.0, "path": 1.0}[
                gradient_normalization
            ]

        keep_irrep_out = self.keep_irrep_out
        if isinstance(keep_irrep_out, str):
            keep_irrep_out = e3nn.Irreps(keep_irrep_out)
            assert all(mul == 1 for mul, _ in keep_irrep_out)

        keep_irrep_out = {e3nn.Irrep(ir) for ir in keep_irrep_out}

        input = input.mul_to_axis().remove_nones()

        ### PREPARE WEIGHTS
        ws = []
        Us = []
        for order in range(1, self.correlation + 1):  # correlation, ..., 1
            if self.symmetric_tensor_product_basis:
                U = e3nn.reduced_symmetric_tensor_product_basis(
                    input.irreps, order, keep_ir=keep_irrep_out
                )
            else:
                U = e3nn.reduced_tensor_product_basis(
                    [input.irreps] * order, keep_ir=keep_irrep_out
                )
            # U = U / order  # normalization TODO(mario): put back after testing
            # NOTE(mario): The normalization constants (/order and /mul**0.5)
            # has been numerically checked to be correct.

            # TODO(mario) implement norm_p
            Us.append(U)

            wsorder = []
            for (mul, ir_out), u in zip(U.irreps, U.list):
                u = u.astype(input.array.dtype)
                # u: ndarray [(irreps_x.dim)^order, multiplicity, ir_out.dim]

                w = self.param(
                    f"w{order}_{ir_out}",
                    nn.initializers.normal(
                        stddev=(mul**-0.5) ** (1.0 - gradient_normalization)
                    ),
                    (self.num_species, mul, input.shape[-2]),
                )
                w = w * (mul**-0.5) ** gradient_normalization
                wsorder.append(w)
            ws.append(wsorder)

        def fn(input: e3nn.IrrepsArray, index: jnp.ndarray):
            # - This operation is parallel on the feature dimension (but each feature has its own parameters)
            # This operation is an efficient implementation of
            # vmap(lambda w, x: FunctionalLinear(irreps_out)(w, concatenate([x, tensor_product(x, x), tensor_product(x, x, x), ...])))(w, x)
            # up to x power self.correlation
            assert input.ndim == 2  # [num_features, irreps_x.dim]
            assert index.ndim == 0  # int

            out = dict()
            x_ = input.array

            for order in range(self.correlation, 0, -1):  # correlation, ..., 1

                U = Us[order - 1]

                # ((w3 x + w2) x + w1) x
                #  \-----------/
                #       out

                for ii, ((mul, ir_out), u) in enumerate(zip(U.irreps, U.list)):
                    u = u.astype(x_.dtype)
                    # u: ndarray [(irreps_x.dim)^order, multiplicity, ir_out.dim]

                    w = ws[order - 1][ii][index]
                    if ir_out not in out:
                        out[ir_out] = (
                            "special",
                            jnp.einsum("...jki,kc,cj->c...i", u, w, x_),
                        )  # [num_features, (irreps_x.dim)^(oder-1), ir_out.dim]
                    else:
                        out[ir_out] += jnp.einsum(
                            "...ki,kc->c...i", u, w
                        )  # [num_features, (irreps_x.dim)^order, ir_out.dim]

                # ((w3 x + w2) x + w1) x
                #  \----------------/
                #         out (in the normal case)

                for ir_out in out:
                    if isinstance(out[ir_out], tuple):
                        out[ir_out] = out[ir_out][1]
                        continue  # already done (special case optimization above)

                    out[ir_out] = jnp.einsum(
                        "c...ji,cj->c...i", out[ir_out], x_
                    )  # [num_features, (irreps_x.dim)^(oder-1), ir_out.dim]

                # ((w3 x + w2) x + w1) x
                #  \-------------------/
                #           out

            # out[irrep_out] : [num_features, ir_out.dim]
            irreps_out = e3nn.Irreps(sorted(out.keys()))
            return e3nn.IrrepsArray.from_list(
                irreps_out,
                [out[ir][:, None, :] for (_, ir) in irreps_out],
                (input.shape[0],),
            )

        # Treat batch indices using vmap
        shape = jnp.broadcast_shapes(input.shape[:-2], index.shape)
        input = input.broadcast_to(shape + input.shape[-2:])
        index = jnp.broadcast_to(index, shape)

        fn_mapped = fn
        for _ in range(input.ndim - 2):
            fn_mapped = jax.vmap(fn_mapped)

        return fn_mapped(input, index).axis_to_mul()


# class SymmetricContraction(nn.Module):

#     correlation: int
#     keep_irrep_out: Set[Irrep]
#     num_species: int
#     gradient_normalization: Union[str, float]
#     symmetric_tensor_product_basis: bool

#     @nn.compact
#     def __call__(self, input: IrrepsArray, index: jnp.ndarray):
#         if not E3NN_AVAILABLE:
#             raise E3NN_EXCEPTION

#         if self.gradient_normalization is None:
#             gradient_normalization = e3nn.config("gradient_normalization")
#         else:
#             gradient_normalization = self.gradient_normalization
#         if isinstance(gradient_normalization, str):
#             gradient_normalization = {"element": 0.0, "path": 1.0}[
#                 gradient_normalization
#             ]

#         keep_irrep_out = self.keep_irrep_out
#         if isinstance(keep_irrep_out, str):
#             keep_irrep_out = e3nn.Irreps(keep_irrep_out)
#             assert all(mul == 1 for mul, _ in keep_irrep_out)

#         keep_irrep_out = {e3nn.Irrep(ir) for ir in keep_irrep_out}

#         onehot = jnp.eye(self.num_species)[index]

#         ### PREPARE WEIGHTS
#         ws = []
#         us = []
#         for ir_out in keep_irrep_out:
#             usorder = []
#             wsorder = []
#             for order in range(1, self.correlation + 1):  # correlation, ..., 1
#                 if self.symmetric_tensor_product_basis:
#                     U = e3nn.reduced_symmetric_tensor_product_basis(
#                         input.irreps, order, keep_ir=[ir_out]
#                     )
#                 else:
#                     U = e3nn.reduced_tensor_product_basis(
#                         [input.irreps] * order, keep_ir=[ir_out]
#                     )
#                 u = jnp.moveaxis(U.list[0].astype(input.array.dtype), -1, 0)
#                 usorder.append(u)

#                 mul, _ = U.irreps[0]
#                 w = self.param(
#                     f"w{order}_{ir_out}",
#                     nn.initializers.normal(
#                         stddev=(mul**-0.5) ** (1.0 - gradient_normalization)
#                     ),
#                     (self.num_species, mul, input.shape[-2]),
#                 )
#                 w = w * (mul**-0.5) ** gradient_normalization
#                 wsorder.append(w)
#             ws.append(wsorder)
#             us.append(usorder)

#         x = input.array

#         outs = []
#         for i, ir in enumerate(keep_irrep_out):
#             w = ws[i][-1]  # [index]
#             u = us[i][-1]
#             out = jnp.einsum("...jk,ekc,bcj,be->bc...", u, w, x, onehot)

#             for order in range(self.correlation - 1, 0, -1):
#                 w = ws[i][order - 1]  # [index]
#                 u = us[i][order - 1]

#                 c_tensor = jnp.einsum("...k,ekc,be->bc...", u, w, onehot) + out
#                 out = jnp.einsum("bc...j,bcj->bc...", c_tensor, x)

#             outs.append(out.reshape(x.shape[0], -1))

#         out = jnp.concatenate(outs, axis=-1)

#         return e3nn.IrrepsArray(input.shape[1] * e3nn.Irreps(keep_irrep_out), out)


import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import  Dict, Optional, ClassVar


class FOAMEmbedding(nn.Module):
    """Filtered Overlap of Atomic Moments
    
    FID : FOAM

    Similar to SOAP embedding but for each rank l, we do not take all combinations
    of each channels but linearly project on 2 nchannels elements and then take the
    scalar product. This is then kind of a linearly filtered SOAP embedding.

    """
    _graphs_properties: Dict
    lmax: int = 2
    """The maximum order of spherical tensors."""
    nchannels: Optional[int] = None
    """The number of channels."""
    graph_key: str = "graph"
    """The key for the graph input."""
    embedding_key: str = "embedding"
    """The key for the embedding output."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`"""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`"""
    include_species: bool = True
    """Whether to include the species encoding in the embedding."""

    FID: ClassVar[str] = "FOAM"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        switch = graph["switch"][:, None]
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_basis = (
            RadialBasis(**{**self.radial_basis, "end": cutoff, "name": "RadialBasis"})(
                graph["distances"]
            )
            * switch
        )

        species_encoding = SpeciesEncoding(
            **self.species_encoding, name="SpeciesEncoding"
        )(species)

        Dij = (radial_basis[:, :, None] * species_encoding[edge_dst, None, :]).reshape(
            -1, species_encoding.shape[-1] * radial_basis.shape[-1]
        )

        Yij = generate_spherical_harmonics(lmax=self.lmax, normalize=False)(
            graph["vec"] / graph["distances"][:, None]
        )[:, :, None]

        rhoi = jax.ops.segment_sum(Dij[:, None, :] * Yij, edge_src, species.shape[0])

        nbasis = rhoi.shape[-1]
        nchannels = self.nchannels if self.nchannels is not None else nbasis

        if self.include_species:
            xis = [species_encoding, rhoi[:, 0, :]]
        else:
            xis = [rhoi[:, 0, :]]

        for l in range(self.lmax + 1):
            rhoil = rhoi[:, l**2 : (l + 1) ** 2, :]
            xl, yl = jnp.split(
                nn.Dense(2 * nchannels, use_bias=False, name=f"xy_l{l}")(rhoil),
                2,
                axis=-1,
            )
            xil = (xl*yl).sum(axis=1) / (2 * l + 1) ** 0.5
            xis.append(xil)
        xi = jnp.concatenate(xis, axis=-1)

        if self.embedding_key is None:
            return xi
        return {**inputs, self.embedding_key: xi}



import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics, CG_SO3
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Dict, Union, Callable, Sequence, Optional, ClassVar
from ...utils.initializers import initializer_from_str, scaled_orthogonal
from ..misc.nets import  ResMLP


class SpookyNetEmbedding(nn.Module):
    """SpookyNet equivariant message-passing embedding with electronic encodings (charge + spin).

    FID : SPOOKYNET

    ### Reference
    Unke, O.T., Chmiela, S., Gastegger, M. et al. SpookyNet: Learning force fields with electronic degrees of freedom and nonlocal effects. Nat Commun 12, 7273 (2021). https://doi.org/10.1038/s41467-021-27504-0

    
    ### Warning
    non-local attention interaction is not yet implemented !

    """

    _graphs_properties: Dict
    dim: int = 128
    """ The dimension of the embedding."""
    nlayers: int = 3
    """ The number of interaction layers."""
    graph_key: str = "graph"
    """ The key for the graph input."""
    embedding_key: str = "embedding"
    """ The key for the embedding output."""
    species_encoding: dict = dataclasses.field(
        default_factory=lambda: {"encoding": "electronic_structure"}
    )
    """ The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`. """
    radial_basis: dict = dataclasses.field(default_factory=lambda: {"basis": "spooky"})
    """ The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`. """
    kernel_init: Union[Callable, str] = "scaled_orthogonal(scale=1.0, mode='fan_avg')"
    """ The kernel initializer for Dense operations."""
    use_spin_encoding: bool = True
    """ Whether to use spin encoding."""
    use_charge_encoding: bool = True
    """ Whether to use charge encoding."""
    total_charge_key: str = "total_charge"
    """ The key for the total charge input."""

    FID: ClassVar[str] = "SPOOKYNET"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]

        cutoff = self._graphs_properties[self.graph_key]["cutoff"]

        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )

        onehot = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(
            species
        )
        zrand = SpeciesEncoding(
            encoding="random", dim=self.dim, name="RandSpeciesEncoding"
        )(species)

        eZ = (
            nn.Dense(
                self.dim,
                name="species_linear",
                use_bias=False,
                kernel_init=nn.initializers.zeros,
            )(onehot)
            + zrand
        )
        xi = eZ

        # encode charge information
        batch_index = inputs["batch_index"]
        natoms = inputs["natoms"]
        if self.use_charge_encoding and (
            self.total_charge_key in inputs or self.is_initializing()
        ):
            Q = inputs.get(self.total_charge_key, jnp.zeros(natoms.shape[0], dtype=xi.dtype))
            kq_pos, kq_neg, vq_pos, vq_neg = self.param(
                "kv_charge",
                lambda key, shape: jax.random.normal(key, shape, dtype=xi.dtype),
                (4, self.dim),
            )
            qi = nn.Dense(self.dim, kernel_init=kernel_init, name="q_linear")(eZ)
            pos_mask = Q >= 0
            kq = jnp.where(pos_mask[:,None], kq_pos[None, :], kq_neg[None, :])
            vq = jnp.where(pos_mask[:,None], vq_pos[None, :], vq_neg[None, :])
            qik = (qi * kq[batch_index]).sum(axis=-1) / self.dim**0.5
            wi =  jax.nn.softplus(qik)
            wnorm = jax.ops.segment_sum(wi, batch_index, Q.shape[0])
            avi = wi[:,None] * ((Q / wnorm)[:, None] * vq)[batch_index]
            eQ = ResMLP(use_bias=False, name="eQ", kernel_init=kernel_init)(avi)
            xi = xi + eQ

        # encode spin information
        if self.use_spin_encoding and (
            "total_spin" in inputs or self.is_initializing()
        ):
            S = inputs.get("total_spin", jnp.zeros(natoms.shape[0], dtype=xi.dtype))
            ks, vs = self.param(
                "kv_spin",
                lambda key, shape: jax.random.normal(key, shape, dtype=xi.dtype),
                (2, self.dim),
            )
            si = nn.Dense(self.dim, kernel_init=kernel_init, name="s_linear")(eZ)
            sik = (si * ks[None, :]).sum(axis=-1) / self.dim**0.5
            wi = jax.nn.softplus(sik)
            wnorm = jax.ops.segment_sum(wi, batch_index, S.shape[0])
            avi = wi[:,None] * ((S / wnorm)[:, None] * vs[None, :])[batch_index]
            eS = ResMLP(use_bias=False, name="eS", kernel_init=kernel_init)(avi)
            xi = xi + eS

        distances = graph["distances"]
        switch = graph["switch"][:, None]
        dirij = graph["vec"] / distances[:, None]
        Yij = generate_spherical_harmonics(lmax=2, normalize=False)(dirij)[:, None, :]

        radial_basis = (
            RadialBasis(
                **{
                    **self.radial_basis,
                    "end": cutoff,
                    "name": f"RadialBasis",
                }
            )(distances)
            * switch
        )

        gij = radial_basis[:, :, None] * Yij

        gsij = gij[:, :, 0]
        gpij = jnp.transpose(gij[:, :, 1:4], (0, 2, 1))
        gdij = jnp.transpose(gij[:, :, 4:], (0, 2, 1))

        y = 0.0

        for layer in range(self.nlayers):
            xtilde = ResMLP(res_only=True, name=f"xtilde_{layer}")(xi)

            ### compute local update
            c = ResMLP(name=f"c_{layer}", kernel_init=kernel_init)(xtilde)
            sj = ResMLP(name=f"s_{layer}", kernel_init=kernel_init)(xtilde)
            pj = ResMLP(name=f"p_{layer}", kernel_init=kernel_init)(xtilde)
            dj = ResMLP(name=f"d_{layer}", kernel_init=kernel_init)(xtilde)

            Gs = nn.Dense(
                self.dim, use_bias=False, name=f"Gs_{layer}", kernel_init=kernel_init
            )(gsij)
            Gp = nn.Dense(
                self.dim, use_bias=False, name=f"Gp_{layer}", kernel_init=kernel_init
            )(gpij)
            Gd = nn.Dense(
                self.dim, use_bias=False, name=f"Gd_{layer}", kernel_init=kernel_init
            )(gdij)

            si = jax.ops.segment_sum(sj[edge_dst] * Gs, edge_src, xi.shape[0])
            pi = jax.ops.segment_sum(pj[edge_dst, None, :] * Gp, edge_src, xi.shape[0])
            di = jax.ops.segment_sum(dj[edge_dst, None, :] * Gd, edge_src, xi.shape[0])

            P1, P2 = jnp.split(
                nn.Dense(
                    2 * self.dim,
                    use_bias=False,
                    name=f"P12_{layer}",
                    kernel_init=kernel_init,
                )(pi),
                2,
                axis=-1,
            )
            D1, D2 = jnp.split(
                nn.Dense(
                    2 * self.dim,
                    use_bias=False,
                    name=f"D12_{layer}",
                    kernel_init=kernel_init,
                )(di),
                2,
                axis=-1,
            )

            P12 = (P1 * P2).sum(axis=1)
            D12 = (D1 * D2).sum(axis=1)

            l = ResMLP(name=f"l_{layer}", kernel_init=kernel_init)(c + si + P12 + D12)

            ### aggregate and update
            xi = ResMLP(name=f"xi_{layer}", kernel_init=kernel_init)(xtilde + l)
            y = y + ResMLP(name=f"y_{layer}", kernel_init=kernel_init)(xi)

        output = {
            **inputs,
            self.embedding_key: y,
        }
        if self.use_charge_encoding and "total_charge" in inputs:
            output[self.embedding_key + "_eQ"] = eQ
        if self.use_spin_encoding and "total_spin" in inputs:
            output[self.embedding_key + "_eS"] = eS
        return output



from .spherical_harmonics import CG_SO3, generate_spherical_harmonics
from .atomic_units import AtomicUnits
from typing import Dict, Any
import jax
import jax.numpy as jnp


def minmaxone(x, name=""):
    print(name, x.min(), x.max(), (x**2).mean() ** 0.5)

def minmaxone_jax(x, name=""):
    jax.debug.print(
        "{name}  {min}  {max}  {mean}",
        name=name,
        min=x.min(),
        max=x.max(),
        mean=(x**2).mean(),
    )


def mask_filter_1d(mask, max_size, *values_fill):
    cumsum = jnp.cumsum(mask,dtype=jnp.int32)
    scatter_idx = jnp.where(mask, cumsum - 1, max_size)
    outputs = []
    for value, fill in values_fill:
        shape = list(value.shape)
        shape[0] = max_size
        output = (
            jnp.full(shape, fill, dtype=value.dtype)
            .at[scatter_idx]
            .set(value, mode="drop")
        )
        outputs.append(output)
    if cumsum.size == 0:
        return outputs, scatter_idx, 0
    return outputs, scatter_idx, cumsum[-1]


def deep_update(
    mapping: Dict[Any, Any], *updating_mappings: Dict[Any, Any]
) -> Dict[Any, Any]:
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping


class Counter:
    def __init__(self, nseg, startsave=1):
        self.i = 0
        self.i_avg = 0
        self.nseg = nseg
        self.startsave = startsave

    @property
    def count(self):
        return self.i

    @property
    def count_avg(self):
        return self.i_avg

    @property
    def nsample(self):
        return max(self.count_avg - self.startsave + 1, 1)

    @property
    def is_reset_step(self):
        return self.count == 0

    def reset_avg(self):
        self.i_avg = 0

    def reset_all(self):
        self.i = 0
        self.i_avg = 0

    def increment(self):
        self.i = self.i + 1
        if self.i >= self.nseg:
            self.i = 0
            self.i_avg = self.i_avg + 1



import jax
import jax.numpy as jnp
import numpy as np
from typing import Callable, Optional, Dict, List, Tuple, Union, Any, NamedTuple
import optax
from copy import deepcopy
from functools import partial

from ..utils import deep_update, AtomicUnits as au
from ..models import FENNIX
        

@partial(jax.jit, static_argnums=(1,2,3,4))
def linear_schedule(step, start_value,end_value, start_step, duration):
    return start_value + jnp.clip((step - start_step) / duration, 0.0, 1.0) * (end_value - start_value)


def get_training_parameters(
    parameters: Dict[str, any], stage: int = -1
) -> Dict[str, any]:
    params = deepcopy(parameters["training"])
    if "stages" not in params:
        return params

    stages: dict = params.pop("stages")
    stage_keys = list(stages.keys())
    if stage < 0:
        stage = len(stage_keys) + stage
    assert stage >= 0 and stage < len(
        stage_keys
    ), f"Stage {stage} not found in training parameters"
    for i in range(stage + 1):
        ## remove end_event from previous stage ##
        if i > 0 and "end_event" in params:
            params.pop("end_event")
        ## incrementally update training parameters ##
        stage_params = stages[stage_keys[i]]
        params = deep_update(params, stage_params)
    return params


def get_loss_definition(
    training_parameters: Dict[str, any],
    model_energy_unit: str = "Ha",  # , manual_renames: List[str] = []
) -> Tuple[Dict[str, any], List[str], List[str]]:
    """
    Returns the loss definition and a list of renamed references.

    Args:
        training_parameters (dict): A dictionary containing training parameters.

    Returns:
        tuple: A tuple containing:
            - loss_definition (dict): A dictionary containing the loss definition.
            - rename_refs (list): A list of renamed references.
    """
    default_loss_type = training_parameters.get("default_loss_type", "log_cosh")
    # loss_definition = deepcopy(training_parameters["loss"])
    used_keys = []
    ref_keys = []
    energy_mult = au.get_multiplier(model_energy_unit)
    loss_definition = {}
    for k in training_parameters["loss"]:
        # loss_prms = loss_definition[k]
        loss_prms = deepcopy(training_parameters["loss"][k])
        if "energy_unit" in loss_prms:
            loss_prms["mult"] = energy_mult / au.get_multiplier(
                loss_prms["energy_unit"]
            )
            if "unit" in loss_prms:
                print(
                    "Warning: Both 'unit' and 'energy_unit' are defined for loss component",
                    k,
                    " -> using 'energy_unit'",
                )
            loss_prms["unit"] = loss_prms["energy_unit"]
        elif "unit" in loss_prms:
            loss_prms["mult"] = 1.0 / au.get_multiplier(loss_prms["unit"])
        else:
            loss_prms["mult"] = 1.0
        if "key" not in loss_prms:
            loss_prms["key"] = k
        if "type" not in loss_prms:
            loss_prms["type"] = default_loss_type
        if "weight" not in loss_prms:
            loss_prms["weight"] = 1.0
        assert loss_prms["weight"] >= 0.0, "Loss weight must be positive"
        if "threshold" in loss_prms:
            assert loss_prms["threshold"] > 1.0, "Threshold must be greater than 1.0"
        if "ref" in loss_prms:
            ref = loss_prms["ref"]
            if not (ref.startswith("model_ref/") or ref.startswith("model/")):
                ref_keys.append(ref)
        if "ds_weight" in loss_prms:
            ref_keys.append(loss_prms["ds_weight"])

        if "weight_start" in loss_prms:
            weight_start = loss_prms["weight_start"]
            if "weight_ramp" in loss_prms:
                weight_ramp = loss_prms["weight_ramp"]
            else:
                weight_ramp = training_parameters.get("max_epochs")
            weight_ramp_start = loss_prms.get("weight_ramp_start", 0.0)
            weight_end = loss_prms["weight"]
            print(f"Weight ramp for {k} : {weight_start} -> {weight_end} in {weight_ramp} epochs")
            loss_prms["weight_schedule"] = (weight_start,weight_end,weight_ramp_start,weight_ramp)

        used_keys.append(loss_prms["key"])
        loss_definition[k] = loss_prms

    # rename_refs = list(
    #     set(["forces", "total_energy", "atomic_energies"] + manual_renames + used_keys)
    # )

    # for k in loss_definition.keys():
    #     loss_prms = loss_definition[k]
    #     if "ref" in loss_prms:
    #         if loss_prms["ref"] in rename_refs:
    #             loss_prms["ref"] = "true_" + loss_prms["ref"]

    return loss_definition, list(set(used_keys)), list(set(ref_keys))


def get_train_step_function(
    loss_definition: Dict,
    model: FENNIX,
    evaluate: Callable,
    optimizer: optax.GradientTransformation,
    ema: Optional[optax.GradientTransformation] = None,
    model_ref: Optional[FENNIX] = None,
    compute_ref_coords: bool = False,
    jit: bool = True,
):
    def train_step(
        data,
        inputs,
        training_state,
    ):
        epoch = training_state["epoch"]

        def loss_fn(variables):
            if model_ref is not None:
                output_ref = evaluate(model_ref, model_ref.variables, inputs)
                # _, _, output_ref = model_ref._energy_and_forces(model_ref.variables, data)
            output = evaluate(model, variables, inputs)
            # _, _, output = model._energy_and_forces(variables, data)
            natoms = jnp.where(inputs["true_sys"], inputs["natoms"], 1)
            nsys = inputs["natoms"].shape[0]
            system_mask = inputs["true_sys"]
            atom_mask = inputs["true_atoms"]
            batch_index = inputs["batch_index"]
            if "system_sign" in inputs:
                system_sign = inputs["system_sign"] > 0
                system_mask = jnp.logical_and(system_mask,system_sign)
                atom_mask = jnp.logical_and(atom_mask, system_sign[batch_index])

            loss_tot = 0.0
            for loss_prms in loss_definition.values():
                use_ref_mask = False
                predicted = output[loss_prms["key"]]
                

                if "ref" in loss_prms:
                    if loss_prms["ref"].startswith("model_ref/"):
                        assert model_ref is not None, "model_ref must be provided"
                        try:
                            ref = output_ref[loss_prms["ref"][10:]] * loss_prms["mult"]
                        except KeyError:
                            raise KeyError(
                                f"Reference key '{loss_prms['ref'][10:]}' not found in model_ref output. Keys available: {output_ref.keys()}"
                            )
                        mask_key  = loss_prms["ref"][6:] + "_mask" 
                        if mask_key in data:
                            use_ref_mask = True
                            ref_mask = data[mask_key]
                        elif mask_key in output_ref:
                            use_ref_mask = True
                            ref_mask = output[mask_key]
                        elif mask_key in output:
                            use_ref_mask = True
                            ref_mask = output[mask_key]
                    elif loss_prms["ref"].startswith("model/"):
                        try:
                            ref = output[loss_prms["ref"][6:]] * loss_prms["mult"]
                        except KeyError:
                            raise KeyError(
                                f"Reference key '{loss_prms['ref'][6:]}' not found in model output. Keys available: {output.keys()}"
                            )
                        mask_key  = loss_prms["ref"][6:] + "_mask" 
                        if mask_key in data:
                            use_ref_mask = True
                            ref_mask = data[mask_key]
                        elif mask_key in output:
                            use_ref_mask = True
                            ref_mask = output[mask_key]
                    else:
                        try:
                            ref = data[loss_prms["ref"]] * loss_prms["mult"]
                        except KeyError:
                            raise KeyError(
                                f"Reference key '{loss_prms['ref']}' not found in data. Keys available: {data.keys()}"
                            )
                        if loss_prms["ref"] + "_mask" in data:
                            use_ref_mask = True
                            ref_mask = data[loss_prms["ref"] + "_mask"]
                else:
                    ref = jnp.zeros_like(predicted)

                if "norm_axis" in loss_prms and loss_prms["norm_axis"] is not None:
                    norm_axis = loss_prms["norm_axis"]
                    predicted = jnp.linalg.norm(
                        predicted, axis=norm_axis, keepdims=True
                    )
                    ref = jnp.linalg.norm(ref, axis=norm_axis, keepdims=True)

                ensemble_loss = loss_prms["type"] in ["ensemble_nll","ensemble_crps","gaussian_mixture",]
                if ensemble_loss:
                    ensemble_axis = loss_prms.get("ensemble_axis", -1)
                    ensemble_weights_key = loss_prms.get("ensemble_weights", None)
                    ensemble_size = predicted.shape[ensemble_axis]
                    if ensemble_weights_key is not None:
                        ensemble_weights = output[ensemble_weights_key]
                    else:
                        ensemble_weights = jnp.ones_like(predicted)
                    if "ensemble_subsample" in loss_prms and "rng_key" in output:
                        ns = min(
                            loss_prms["ensemble_subsample"],
                            predicted.shape[ensemble_axis],
                        )
                        key, subkey = jax.random.split(output["rng_key"])

                        def get_subsample(x):
                            return jax.lax.slice_in_dim(
                                jax.random.permutation(
                                    subkey, x, axis=ensemble_axis, independent=True
                                ),
                                start_index=0,
                                limit_index=ns,
                                axis=ensemble_axis,
                            )

                        predicted = get_subsample(predicted)
                        ensemble_weights = get_subsample(ensemble_weights)
                        output["rng_key"] = key
                    else:
                        get_subsample = lambda x: x
                    ensemble_weights = ensemble_weights / jnp.sum(
                        ensemble_weights, axis=ensemble_axis, keepdims=True
                    )
                    predicted_ensemble = predicted
                    predicted = (predicted * ensemble_weights).sum(
                        axis=ensemble_axis, keepdims=True
                    )
                    predicted_var = (
                        ensemble_weights * (predicted_ensemble - predicted) ** 2
                    ).sum(axis=ensemble_axis) * (ensemble_size / (ensemble_size - 1.0))
                    predicted = jnp.squeeze(predicted, axis=ensemble_axis)
                
                if "remove_ref_sys" in loss_prms and loss_prms["remove_ref_sys"]:
                    assert compute_ref_coords, "compute_ref_coords must be True"
                    # predicted = predicted - output_data_ref[loss_prms["key"]]
                    # assert predicted.shape[0] == nsys, "remove_ref_sys only works with system-level predictions"
                    shape_mask = [predicted.shape[0]] + [1] * (len(predicted.shape) - 1)
                    if predicted.shape[0] == nsys:
                        predicted_sign_mask = inputs["system_sign"].reshape(*shape_mask) < 0
                        predicted_shift = jax.ops.segment_sum(predicted_sign_mask*predicted,inputs["system_index"],nsys)
                        predicted = predicted - predicted_shift
                    elif predicted.shape[0] == batch_index.shape[0]:
                        predicted_sign_mask = (inputs["system_sign"]<0)[batch_index].reshape(*shape_mask)
                        natshift = jnp.concatenate([jnp.array([0]),jnp.cumsum(natoms)])
                        iat = jnp.arange(batch_index.shape[0]) - natshift[batch_index]
                        iatdest = natshift[inputs["system_index"]][batch_index]+iat
                        predicted_shift = jax.ops.segment_sum(predicted_sign_mask*predicted,iatdest,batch_index.shape[0])
                        predicted = predicted - predicted_shift
                    else:
                        raise ValueError("remove_ref_sys only works with system-level or atom-level predictions")
                    
                    if ensemble_loss:
                        predicted_ensemble = predicted_ensemble - jnp.expand_dims(predicted_shift,ensemble_axis)
                        predicted_var = (
                            ensemble_weights * (predicted_ensemble - jnp.expand_dims(predicted,ensemble_axis)) ** 2
                        ).sum(axis=ensemble_axis) * (ensemble_size / (ensemble_size - 1.0))

                if predicted.ndim > 1 and predicted.shape[-1] == 1:
                    predicted = jnp.squeeze(predicted, axis=-1)

                if ref.ndim > 1 and ref.shape[-1] == 1:
                    ref = jnp.squeeze(ref, axis=-1)

                per_atom = False
                shape_mask = [ref.shape[0]] + [1] * (len(ref.shape) - 1)
                # print(loss_prms["key"],predicted.shape,loss_prms["ref"],ref.shape)
                natscale = 1.0
                if ref.shape[0] == output["batch_index"].shape[0]:
                    ## shape is number of atoms
                    truth_mask = atom_mask
                    if "ds_weight" in loss_prms:
                        weight_key = loss_prms["ds_weight"]
                        natscale=data[weight_key]
                        if natscale.shape[0] == natoms.shape[0]:
                            natscale = natscale[output["batch_index"]]
                        natscale = natscale.reshape(*shape_mask)
                elif ref.shape[0] == natoms.shape[0]:
                    ## shape is number of systems
                    truth_mask = system_mask
                    if loss_prms.get("per_atom", False):
                        per_atom = True
                        ref = ref / natoms.reshape(*shape_mask)
                        predicted = predicted / natoms.reshape(*shape_mask)
                    if "nat_pow" in loss_prms:
                        natscale = (
                            1.0 / natoms.reshape(*shape_mask) ** loss_prms["nat_pow"]
                        )
                    if "ds_weight" in loss_prms:
                        weight_key = loss_prms["ds_weight"]
                        natscale = natscale * data[weight_key].reshape(*shape_mask)
                else:
                    truth_mask = jnp.ones(ref.shape[0], dtype=bool)

                if use_ref_mask:
                    truth_mask = truth_mask * ref_mask.astype(bool)

                nel = jnp.maximum(
                    (float(np.prod(ref.shape)) / float(truth_mask.shape[0]))
                    * jnp.sum(truth_mask).astype(jnp.float32),
                    1.0,
                )
                truth_mask = truth_mask.reshape(*shape_mask)

                ref = ref * truth_mask
                predicted = predicted * truth_mask

                natscale = natscale / nel

                loss_type = loss_prms["type"]
                if loss_type == "mse":
                    loss = jnp.sum(natscale * (predicted - ref) ** 2)
                elif loss_type == "log_cosh":
                    loss = jnp.sum(natscale * optax.log_cosh(predicted, ref))
                elif loss_type == "mae":
                    loss = jnp.sum(natscale * jnp.abs(predicted - ref))
                elif loss_type == "rel_mse":
                    eps = loss_prms.get("rel_eps", 1.0e-6)
                    rel_pow = loss_prms.get("rel_pow", 2.0)
                    loss = jnp.sum(
                        natscale
                        * (predicted - ref) ** 2
                        / (eps + jnp.abs(ref) ** rel_pow)
                    )
                elif loss_type == "rmse+mae":
                    loss = (
                        jnp.sum(natscale * (predicted - ref) ** 2)
                    ) ** 0.5 + jnp.sum(natscale * jnp.abs(predicted - ref))
                elif loss_type == "crps":
                    predicted_var_key = loss_prms.get(
                        "var_key", loss_prms["key"] + "_var"
                    )
                    lamba_crps = loss_prms.get("lambda_crps", 1.0)
                    predicted_var = output[predicted_var_key]
                    if per_atom:
                        predicted_var = predicted_var / natoms.reshape(*shape_mask)**2
                    predicted_var = predicted_var * truth_mask + (1.0 - truth_mask)
                    sigma = predicted_var**0.5
                    dy = (ref - predicted) / sigma
                    Phi = 0.5 * (1.0 + jax.scipy.special.erf(dy / 2**0.5))
                    phi = jnp.exp(-0.5 * dy**2) / (2 * jnp.pi) ** 0.5
                    loss = jnp.sum(natscale * sigma * (dy * (2 * Phi - 1.0) + 2 * phi - lamba_crps/jnp.pi**0.5))
                elif loss_type == "ensemble_nll":
                    predicted_var = predicted_var * truth_mask + (1.0 - truth_mask)
                    loss = 0.5 * jnp.sum(
                        natscale
                        * truth_mask
                        * (
                            jnp.log(predicted_var)
                            + (ref - predicted) ** 2 / predicted_var
                        )
                    )
                elif loss_type == "ensemble_crps":
                    if per_atom:
                        predicted_var = predicted_var / natoms.reshape(*shape_mask)**2
                    lamba_crps = loss_prms.get("lambda_crps", 1.0)
                    predicted_var = predicted_var * truth_mask + (1.0 - truth_mask)
                    sigma = predicted_var**0.5
                    dy = (ref - predicted) / sigma
                    Phi = 0.5 * (1.0 + jax.scipy.special.erf(dy / 2**0.5))
                    phi = jnp.exp(-0.5 * dy**2) / (2 * jnp.pi) ** 0.5
                    loss = jnp.sum(
                        natscale * truth_mask * sigma * (dy * (2 * Phi - 1.0) + 2 * phi - lamba_crps/jnp.pi**0.5)
                    )
                elif loss_type == "gaussian_mixture":
                    gm_w = get_subsample(output[loss_prms["key"] + "_w"])
                    gm_w = gm_w / jnp.sum(gm_w, axis=ensemble_axis, keepdims=True)
                    sigma = get_subsample(output[loss_prms["key"] + "_sigma"])
                    ref = jnp.expand_dims(ref, axis=ensemble_axis)
                    # log_likelihoods = -0.5*jnp.log(2*np.pi*gm_sigma2) - ((ref - predicted_ensemble) ** 2) / (2 * gm_sigma2)
                    # log_w =jnp.log(gm_w + 1.e-10)
                    # negloglikelihood = -jax.scipy.special.logsumexp(log_w + log_likelihoods,axis=ensemble_axis)
                    # loss = jnp.sum(natscale * truth_mask * negloglikelihood)

                    # CRPS loss
                    def compute_crps(mu, sigma):
                        dy = mu / sigma
                        Phi = 0.5 * (1.0 + jax.scipy.special.erf(dy / 2**0.5))
                        phi = jnp.exp(-0.5 * dy**2) / (2 * jnp.pi) ** 0.5
                        return sigma * (dy * (2 * Phi - 1.0) + 2 * phi - 1./jnp.pi**0.5)

                    crps1 = compute_crps(ref - predicted_ensemble, sigma)
                    gm_crps1 = (gm_w * crps1).sum(axis=ensemble_axis)

                    if ensemble_axis < 0:
                        ensemble_axis = predicted_ensemble.ndim + ensemble_axis
                    muij = jnp.expand_dims(
                        predicted_ensemble, axis=ensemble_axis
                    ) - jnp.expand_dims(predicted_ensemble, axis=ensemble_axis + 1)
                    sigma2 = sigma**2
                    sigmaij = (
                        jnp.expand_dims(sigma2, axis=ensemble_axis)
                        + jnp.expand_dims(sigma2, axis=ensemble_axis + 1)
                    ) ** 0.5
                    wij = jnp.expand_dims(gm_w, axis=ensemble_axis) * jnp.expand_dims(
                        gm_w, axis=ensemble_axis + 1
                    )
                    crps2 = compute_crps(muij, sigmaij)
                    gm_crps2 = (wij * crps2).sum(
                        axis=(ensemble_axis, ensemble_axis + 1)
                    )

                    crps = gm_crps1 - 0.5 * gm_crps2
                    loss = jnp.sum(natscale * truth_mask * crps)

                elif loss_type == "evidential":
                    evidence = loss_prms["evidence_key"]
                    nu, alpha, beta = jnp.split(output[evidence], 3, axis=-1)
                    gamma = predicted
                    nu = nu.reshape(shape_mask)
                    alpha = alpha.reshape(shape_mask)
                    beta = beta.reshape(shape_mask)
                    nu = jnp.where(truth_mask, nu, 1.0)
                    alpha = jnp.where(truth_mask, alpha, 1.0)
                    beta = jnp.where(truth_mask, beta, 1.0)
                    omega = 2 * beta * (1 + nu)
                    lg = jax.scipy.special.gammaln(alpha) - jax.scipy.special.gammaln(
                        alpha + 0.5
                    )
                    ls = 0.5 * jnp.log(jnp.pi / nu) - alpha * jnp.log(omega)
                    lt = (alpha + 0.5) * jnp.log(omega + nu * (gamma - ref) ** 2)
                    wst = (
                        (beta * (1 + nu) / (alpha * nu)) ** 0.5
                        if loss_prms.get("normalize_evidence", True)
                        else 1.0
                    )
                    lr = (
                        loss_prms.get("lambda_evidence", 1.0)
                        * jnp.abs(gamma - ref)
                        * nu
                        / wst
                    )
                    r = loss_prms.get("evidence_ratio", 1.0)
                    le = (
                        loss_prms.get("lambda_evidence_diff", 0.0)
                        * (nu - r * 2 * alpha) ** 2
                    )
                    lb = loss_prms.get("lambda_evidence_beta", 0.0) * beta
                    loss = lg + ls + lt + lr + le + lb

                    loss = jnp.sum(natscale * loss * truth_mask)
                elif loss_type == "raw":
                    loss = jnp.sum(natscale * predicted)
                elif loss_type == "binary_crossentropy":
                    ref = ref.astype(predicted.dtype)
                    if loss_prms.get("raw_logits",False):
                        loss = jnp.sum(
                            natscale*truth_mask
                            * (
                                -ref * jax.nn.log_sigmoid(predicted)
                                - (1.0 - ref) * jax.nn.log_sigmoid(-predicted)
                            )
                        )
                    else:
                        loss = jnp.sum(
                            natscale*truth_mask
                            * (
                                -ref * jnp.log(predicted + 1.0e-10)
                                - (1.0 - ref) * jnp.log(1.0 - predicted + 1.0e-10)
                            )
                        )
                else:
                    raise ValueError(f"Unknown loss type: {loss_type}")

                if "weight_schedule" in loss_prms:
                    w = linear_schedule(epoch, *loss_prms["weight_schedule"])
                else:
                    w = loss_prms["weight"]

                loss_tot = loss_tot + w * loss

            return loss_tot, output

        variables = training_state["variables"]
        new_training_state = {**training_state, "step":training_state["step"]+1}

        (loss, o), grad = jax.value_and_grad(loss_fn, has_aux=True)(variables)
        updates, new_training_state["opt_state"] = optimizer.update(grad, training_state["opt_state"], params=variables)
        variables = optax.apply_updates(variables, updates)
        new_training_state["variables"] = variables

        if ema is not None:
            ema_st = training_state.get("ema_state",None)
            if ema_st is None:
                raise ValueError(
                    "train_step was setup with ema but either variables_ema or ema_st was not provided"
                )
            new_training_state["variables_ema"], new_training_state["ema_state"] = ema.update(variables, ema_st)
    
        return loss, new_training_state, o

    if jit:
        return jax.jit(train_step)
    return train_step


def get_validation_function(
    loss_definition: Dict,
    model: FENNIX,
    evaluate: Callable,
    model_ref: Optional[FENNIX] = None,
    compute_ref_coords: bool = False,
    return_targets: bool = False,
    jit: bool = True,
):

    def validation(data, inputs, variables, inputs_ref=None):
        if model_ref is not None:
            output_ref = evaluate(model_ref, model_ref.variables, inputs)
            # _, _, output_ref = model_ref._energy_and_forces(model_ref.variables, data)
        output = evaluate(model, variables, inputs)
        # _, _, output = model._energy_and_forces(variables, data)
        rmses = {}
        maes = {}
        if return_targets:
            targets = {}

        natoms = jnp.where(inputs["true_sys"], inputs["natoms"], 1)
        nsys = inputs["natoms"].shape[0]
        system_mask = inputs["true_sys"]
        atom_mask = inputs["true_atoms"]
        batch_index = inputs["batch_index"]
        if "system_sign" in inputs:
            system_sign = inputs["system_sign"] > 0
            system_mask = jnp.logical_and(system_mask,system_sign)
            atom_mask = jnp.logical_and(atom_mask, system_sign[batch_index])

        for name, loss_prms in loss_definition.items():
            do_validation = loss_prms.get("validate", True)
            if not do_validation:
                continue
            predicted = output[loss_prms["key"]]
            use_ref_mask = False
            
            if "remove_ref_sys" in loss_prms and loss_prms["remove_ref_sys"]:
                assert compute_ref_coords, "compute_ref_coords must be True"
                # predicted = predicted - output_data_ref[loss_prms["key"]]
                shape_mask = [predicted.shape[0]] + [1] * (len(predicted.shape) - 1)
                if predicted.shape[0] == nsys:
                    system_sign = inputs["system_sign"].reshape(*shape_mask)
                    predicted = jax.ops.segment_sum(system_sign*predicted,inputs["system_index"],nsys)
                elif predicted.shape[0] == batch_index.shape[0]:
                    atom_sign = inputs["system_sign"][batch_index].reshape(*shape_mask)
                    natshift = jnp.concatenate([jnp.array([0]),jnp.cumsum(natoms)])
                    iat = jnp.arange(batch_index.shape[0]) - natshift[batch_index]
                    iatdest = natshift[inputs["system_index"]][batch_index]+iat
                    predicted = jax.ops.segment_sum(atom_sign*predicted,iatdest,batch_index.shape[0])
                else:
                    raise ValueError("remove_ref_sys only works with system-level or atom-level predictions")

            if "ref" in loss_prms:
                if loss_prms["ref"].startswith("model_ref/"):
                    assert model_ref is not None, "model_ref must be provided"
                    ref = output_ref[loss_prms["ref"][10:]] * loss_prms["mult"]
                    mask_key  = loss_prms["ref"][6:] + "_mask" 
                    if mask_key in data:
                        use_ref_mask = True
                        ref_mask = data[mask_key]
                    elif mask_key in output_ref:
                        use_ref_mask = True
                        ref_mask = output[mask_key]
                    elif mask_key in output:
                        use_ref_mask = True
                        ref_mask = output[mask_key]
                elif loss_prms["ref"].startswith("model/"):
                    ref = output[loss_prms["ref"][6:]] * loss_prms["mult"]
                    mask_key  = loss_prms["ref"][6:] + "_mask" 
                    if mask_key in data:
                        use_ref_mask = True
                        ref_mask = data[mask_key]
                    elif mask_key in output:
                        use_ref_mask = True
                        ref_mask = output[mask_key]
                else:
                    ref = data[loss_prms["ref"]] * loss_prms["mult"]
                    if loss_prms["ref"] + "_mask" in data:
                        use_ref_mask = True
                        ref_mask = data[loss_prms["ref"] + "_mask"]
            else:
                ref = jnp.zeros_like(predicted)
            
            if "norm_axis" in loss_prms and loss_prms["norm_axis"] is not None:
                norm_axis = loss_prms["norm_axis"]
                predicted = jnp.linalg.norm(
                    predicted, axis=norm_axis, keepdims=True
                )
                ref = jnp.linalg.norm(ref, axis=norm_axis, keepdims=True)

            loss_type = loss_prms["type"].lower()
            if loss_type.startswith("ensemble"):
                axis = loss_prms.get("ensemble_axis", -1)
                ensemble_weights_key = loss_prms.get("ensemble_weights", None)
                if ensemble_weights_key is not None:
                    ensemble_weights = output[ensemble_weights_key]
                else:
                    ensemble_weights = jnp.ones_like(predicted)
                ensemble_weights = ensemble_weights / jnp.sum(
                    ensemble_weights, axis=axis, keepdims=True
                )

                # predicted = predicted.mean(axis=axis)
                predicted = (ensemble_weights * predicted).sum(axis=axis)

            if loss_prms["type"] in ["gaussian_mixture"]:
                axis = loss_prms.get("ensemble_axis", -1)
                gm_w = output[loss_prms["key"] + "_w"]
                gm_w = gm_w / jnp.sum(gm_w, axis=axis, keepdims=True)
                predicted = (predicted * gm_w).sum(axis=axis)

            if predicted.ndim > 1 and predicted.shape[-1] == 1:
                predicted = jnp.squeeze(predicted, axis=-1)

            if ref.ndim > 1 and ref.shape[-1] == 1:
                ref = jnp.squeeze(ref, axis=-1)

            shape_mask = [ref.shape[0]] + [1] * (len(predicted.shape) - 1)
            if ref.shape[0] == output["batch_index"].shape[0]:
                ## shape is number of atoms
                truth_mask = atom_mask
            elif ref.shape[0] == natoms.shape[0]:
                ## shape is number of systems
                truth_mask = system_mask
                if loss_prms.get("per_atom_validation", False):
                    ref = ref / natoms.reshape(*shape_mask)
                    predicted = predicted / natoms.reshape(*shape_mask)
            else:
                truth_mask = jnp.ones(ref.shape[0], dtype=bool)

            if use_ref_mask:
                truth_mask = truth_mask * ref_mask.astype(bool)

            nel = jnp.maximum(
                (float(np.prod(ref.shape)) / float(truth_mask.shape[0]))
                * jnp.sum(truth_mask).astype(predicted.dtype),
                1.0,
            )

            truth_mask = truth_mask.reshape(*shape_mask)

            ref = ref * truth_mask
            predicted = predicted * truth_mask

            if loss_type == "binary_crossentropy":
                ref = ref.astype(predicted.dtype)
                if loss_prms.get("raw_logits",False):
                    rmse = jnp.sum(
                            (
                                -ref * jax.nn.log_sigmoid(predicted)
                                - (1.0 - ref) * jax.nn.log_sigmoid(-predicted)
                            )*truth_mask/nel
                        )
                    predicted = truth_mask*jax.nn.sigmoid(predicted)
                else:
                    rmse = jnp.sum(
                            (
                                -ref * jnp.log(predicted + 1.0e-10)
                                - (1.0 - ref) * jnp.log(1.0 - predicted + 1.0e-10)
                            )*truth_mask/nel
                        )
                thr = loss_prms.get("class_threshold", 0.05)
                mae = jnp.sum(jnp.where(jnp.abs(predicted - ref) > thr, 1, 0)/nel)*100
            else:
                rmse = jnp.sum((predicted - ref) ** 2 / nel) ** 0.5
                mae = jnp.sum(jnp.abs(predicted - ref) / nel)

            rmses[name] = rmse
            maes[name] = mae
            if return_targets:
                targets[name] = (predicted, ref, truth_mask)

        if return_targets:
            return rmses, maes, output, targets

        return rmses, maes, output

    if jit:
        return jax.jit(validation)
    return validation



import re
from .atomic_units import AtomicUnits as au
_separators=" |,|=|\t|\n"
_comment_chars=["#","!"]
_true_repr=["true","yes",".true."]
_false_repr=["false","no",".false."]

class InputFile(dict):
	case_insensitive=True
	def get(self, path, default = None):
		if InputFile.case_insensitive:
			path=path.lower()
		keys = path.split("/")
		val = None
		for key in keys:
			if isinstance(val,InputFile):
				val = val.get(key, default=None)
			else:
				val = dict.get(self, key, None)

			if val is None:
				return default
		
		return val

	def store(self, path, value):
		if InputFile.case_insensitive:
			path=path.lower()
		keys = path.split("/")
		child = self.get(keys[0],default=None)
		if isinstance(child,InputFile):
			if len(keys) == 1:
				print("Warning: overriding a sub-dictionary!")
				self[keys[0]]=value
				return 1
			else:
				child.store("/".join(keys[1:]),value)
		else:
			if len(keys) == 1:
				self[keys[0]]=value
				return 0
			else:
				if child is None:
					self[keys[0]]=InputFile()
					self[keys[0]].store("/".join(keys[1:]),value)
				else:
					print("Error: hit a leaf before the end of path!")
					return -1

	
	def print(self,tab=""):
		string = ""
		for p_id, p_info in self.items():
			string += tab+p_id
			val=self.get(p_id)
			if isinstance(val,InputFile):
				string += "{\n" + val.print(tab=tab+"  ") +"\n"+tab+"}\n\n"
			else:
				string += " = "+str(val)+"\n"			
		return string[:-1]
	
	def save(self,filename):
		with open(filename,"w") as f:
			f.write(self.print())
	
	def __str__(self):
		return self.print()


def parse_input(input_file):
# parse an input file and return a nested dictionary
# containing the categories, keys and values
	f=open(input_file,'r')
	struct=InputFile()
	path=[]
	for line in f:
		#remove all after comment character
		for comment_char in _comment_chars:
			index=line.find(comment_char)
			if index >= 0:
				line=line[:index]

		#split line using defined separators
		parsed_line=re.split(_separators,
						line.strip())

		#remove empty strings
		parsed_line=[x for x in parsed_line if x]
		#skip blank lines
		if not parsed_line:
			continue
		#print(parsed_line)

		word0 = parsed_line[0].lower()
		cat_fields=''.join(parsed_line)
		#check if beginning of a category
		if cat_fields.endswith("{"):
			path.append(cat_fields[:-1])
			continue
		if cat_fields.startswith("&"):
			path.append(cat_fields[1:])
			continue
		if cat_fields.endswith("{}"):
			struct.store("/".join("path")+"/"+cat_fields[1:-2]
							,InputFile())
			continue		
		
		#print(current_category)
		#if not path:
		#	print("Error: line not recognized!")
		#	return None
		# else: #check if end of a category
		if (cat_fields[0] in "}/") or ("&end" in cat_fields):
			del path[-1]
			continue
		
		word0, unit=_get_unit_from_key(word0)
		val=None
		if len(parsed_line) == 1:
			val=True  # keyword only => store True
		elif len(parsed_line) == 2:
			val=string_to_true_type(parsed_line[1],unit)
		else:
			#analyze parsed line
			val=[]
			for word in parsed_line[1:]:
				val.append(string_to_true_type(word,unit))
		struct.store("/".join(path+[word0]),val)

	f.close()
	return struct

def string_to_true_type(word,unit=None):
	if unit is not None:
		return float(word)/unit
	
	try:
		val=int(word)
	except ValueError:
		try:
			val=float(word)
		except ValueError:
			if word.lower() in _true_repr:
				val=True
			elif word.lower() in _false_repr:
				val=False
			else:
				val=word
	return val

def _get_unit_from_key(word):
	unit_start=max(word.find("{"),word.find("["))
	n=len(word)
	if unit_start<0:
		key=word
		unit=None
	elif unit_start==0:
		print("Error: Field '"+str(word)+"' must not start with '{' or '[' !")
		raise ValueError
	else:
		if word[unit_start] == "{":
			end_bracket="}"
		else:
			end_bracket="]"
		key=word[:unit_start]
		if word[n-1] != end_bracket:
			print("Error: wrong unit specification in field '"+str(word)+"' !")
			raise ValueError
		
		if n-unit_start-2 < 0 :
			unit=1.
		else :
			unit=au.get_multiplier(word[unit_start+1:-1])
			#print(key+" unit= "+str(unit))
	return key, unit



import flax.linen as nn
from typing import Any, Optional, ClassVar
import jax.numpy as jnp
import jax
import numpy as np
from ...utils.spherical_harmonics import CG_SO3, spherical_to_cartesian_tensor

### e3nn version
try:
    import e3nn_jax as e3nn
    E3NN_AVAILABLE = True
    E3NN_EXCEPTION = None
    Irreps = e3nn.Irreps
except Exception as e:
    E3NN_AVAILABLE = False
    E3NN_EXCEPTION = e
    class Irreps(tuple):
        pass


class FullTensorProduct(nn.Module):
    """Tensor product of two spherical harmonics."""

    lmax1: int
    """The maximum order of the first spherical tensor."""
    lmax2: int
    """The maximum order of the second spherical tensor."""
    lmax_out: Optional[int] = None
    """The maximum order of the output spherical tensor. If None, it is the sum of lmax1 and lmax2."""
    ignore_parity: bool = False
    """Whether to ignore the parity of the spherical tensors."""

    @nn.compact
    def __call__(self, x1, x2) -> None:
        irreps_1 = [(l, (-1) ** l) for l in range(self.lmax1 + 1)]
        irreps_2 = [(l, (-1) ** l) for l in range(self.lmax2 + 1)]
        irreps_out = []
        lsout = []
        psout = []
        i12 = []

        lmax_out = self.lmax_out or self.lmax1 + self.lmax2

        for i1, (l1, p1) in enumerate(irreps_1):
            for i2, (l2, p2) in enumerate(irreps_2):
                for lout in range(abs(l1 - l2), l1 + l2 + 1):
                    if p1 * p2 != (-1) ** lout and not self.ignore_parity:
                        continue
                    if lout > lmax_out:
                        continue
                    lsout.append(lout)
                    psout.append(p1 * p2)
                    i12.append((i1, i2))

        lsout = np.array(lsout)
        psout = np.array(psout)
        idx = np.lexsort((psout, lsout))
        lsout = lsout[idx]
        psout = psout[idx]
        i12 = [i12[i] for i in idx]
        irreps_out = [(l, p) for l, p in zip(lsout, psout)]

        slices_1 = [0]
        for l, p in irreps_1:
            slices_1.append(slices_1[-1] + 2 * l + 1)
        slices_2 = [0]
        for l, p in irreps_2:
            slices_2.append(slices_2[-1] + 2 * l + 1)
        slices_out = [0]
        for l, p in irreps_out:
            slices_out.append(slices_out[-1] + 2 * l + 1)

        assert slices_1[-1] == (self.lmax1 + 1) ** 2
        assert slices_2[-1] == (self.lmax2 + 1) ** 2

        shape = ((self.lmax1 + 1) ** 2, (self.lmax2 + 1) ** 2, slices_out[-1])
        w3js = np.zeros(shape)
        for iout, (lout, pout) in enumerate(irreps_out):
            i1, i2 = i12[iout]
            l1, p1 = irreps_1[i1]
            l2, p2 = irreps_2[i2]
            w3j = CG_SO3(l1, l2, lout)
            scale = (2 * lout + 1) ** 0.5
            w3js[
                slices_1[i1] : slices_1[i1 + 1],
                slices_2[i2] : slices_2[i2 + 1],
                slices_out[iout] : slices_out[iout + 1],
            ] = (
                w3j * scale
            )
        w3j = jnp.asarray(w3js)

        return jnp.einsum("...a,...b,abc->...c", x1, x2, w3j), irreps_out


class FilteredTensorProduct(nn.Module):
    """Tensor product of two spherical harmonics filtered to give back the irreps of the first input"""

    lmax1: int
    """The maximum order of the first spherical tensor."""
    lmax2: int
    """The maximum order of the second spherical tensor."""
    lmax_out: Optional[int] = None
    """The maximum order of the output spherical tensor. If None, it is the same as lmax1."""
    ignore_parity: bool = False
    """Whether to ignore the parity of the spherical tensors."""
    weights_by_channel: bool = False
    """Whether to use different path weights for each channel."""

    @nn.compact
    def __call__(self, x1, x2) -> None:
        irreps_1 = [(l, (-1) ** l) for l in range(self.lmax1 + 1)]
        irreps_2 = [(l, (-1) ** l) for l in range(self.lmax2 + 1)]
        lmax_out = self.lmax_out if self.lmax_out is not None else self.lmax1
        irreps_out = [(l, (-1) ** l) for l in range(lmax_out + 1)]

        slices_1 = [0]
        for l, p in irreps_1:
            slices_1.append(slices_1[-1] + 2 * l + 1)
        slices_2 = [0]
        for l, p in irreps_2:
            slices_2.append(slices_2[-1] + 2 * l + 1)
        slices_out = [0]
        for l, p in irreps_out:
            slices_out.append(slices_out[-1] + 2 * l + 1)

        shape = (x1.shape[-1], x2.shape[-1], x1.shape[-1])
        w3js = []
        for iout, (lout, pout) in enumerate(irreps_out):
            for i1, (l1, p1) in enumerate(irreps_1):
                for i2, (l2, p2) in enumerate(irreps_2):
                    if pout != p1 * p2 and not self.ignore_parity:
                        continue
                    if lout > l1 + l2 or lout < abs(l1 - l2):
                        continue
                    w3j = CG_SO3(l1, l2, lout)
                    w3j_full = np.zeros(shape)
                    scale = (2 * lout + 1) ** 0.5
                    w3j_full[
                        slices_1[i1] : slices_1[i1 + 1],
                        slices_2[i2] : slices_2[i2 + 1],
                        slices_out[iout] : slices_out[iout + 1],
                    ] = (
                        w3j * scale
                    )
                    w3js.append(w3j_full)
        npath = len(w3js)
        w3j = jnp.asarray(np.stack(w3js))

        if self.weights_by_channel:
            nchannels = x1.shape[-2]
            weights = self.param("weights", jax.nn.initializers.normal(stddev=1./npath**0.5), (nchannels,npath,))
            ww3j = jnp.einsum("np,pabc->nabc", weights, w3j)
            return jnp.einsum("...na,...nb,nabc->...nc", x1, x2, ww3j)


        weights = self.param("weights", jax.nn.initializers.normal(stddev=1./npath**0.5), (npath,))

        ww3j = jnp.einsum("p,pabc->abc", weights, w3j)
        return jnp.einsum("...a,...b,abc->...c", x1, x2, ww3j)


class ChannelMixing(nn.Module):
    """Linear mixing of input channels.
    
    FID: CHANNEL_MIXING
    """

    lmax: int
    """The maximum order of the spherical tensor."""
    nchannels: int
    """The number of input channels."""
    nchannels_out: Optional[int] = None
    """The number of output channels. If None, it is the same as nchannels."""
    input_key: Optional[str] = None
    """The key in the input dictionary that corresponds to the input tensor."""
    output_key: Optional[str] = None
    """The key in the output dictionary where the computed tensor will be stored."""
    squeeze: bool = False
    """Whether to squeeze the output tensor to remove the channel dimension."""

    FID: ClassVar[str] = "CHANNEL_MIXING"

    @nn.compact
    def __call__(self, inputs):
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        ########################################
        nchannels_out = self.nchannels_out or self.nchannels
        weights = self.param(
            "weights",
            jax.nn.initializers.normal(stddev=1./self.nchannels**0.5),
            (nchannels_out, self.nchannels),
        )
        out = jnp.einsum("ij,...jk->...ik", weights, x)
        if self.squeeze and nchannels_out == 1:
            out = jnp.squeeze(out, axis=-2)
        ########################################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class ChannelMixingE3(nn.Module):
    """Linear mixing of input channels with different weight for each angular momentum.
    
    FID: CHANNEL_MIXING_E3
    """

    lmax: int
    """The maximum order of the spherical tensor."""
    nchannels: int
    """The number of input channels."""
    nchannels_out: Optional[int] = None
    """The number of output channels. If None, it is the same as nchannels."""
    input_key: Optional[str] = None
    """The key in the input dictionary that corresponds to the input tensor."""
    output_key: Optional[str] = None
    """The key in the output dictionary where the computed tensor will be stored."""
    squeeze: bool = False
    """Whether to squeeze the output tensor to remove the channel dimension."""

    FID: ClassVar[str] = "CHANNEL_MIXING_E3"

    @nn.compact
    def __call__(self, inputs):
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        ########################################
        nrep = np.array([2 * l + 1 for l in range(self.lmax + 1)])
        nchannels_out = self.nchannels_out or self.nchannels
        weights = jnp.repeat(
            self.param(
                "weights",
                jax.nn.initializers.normal(stddev=1./self.nchannels**0.5),
                (nchannels_out, self.nchannels, self.lmax + 1),
            ),
            nrep,
            axis=-1,
        )
        out = jnp.einsum("ijk,...jk->...ik", weights, x)
        if self.squeeze and nchannels_out == 1:
            out = jnp.squeeze(out, axis=-2)
        ########################################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class SphericalToCartesian(nn.Module):
    """Convert spherical tensors to cartesian tensors.
    
    FID: SPHERICAL_TO_CARTESIAN
    """
    
    lmax: int
    """The maximum order of the spherical tensor. Only implemented for lmax up to 2."""
    input_key: Optional[str] = None
    """The key in the input dictionary that corresponds to the input tensor."""
    output_key: Optional[str] = None
    """The key in the output dictionary where the computed tensor will be stored."""

    FID: ClassVar[str] = "SPHERICAL_TO_CARTESIAN"

    @nn.compact
    def __call__(self, inputs) -> Any:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        ########################################
        out = spherical_to_cartesian_tensor(x, self.lmax)
        ########################################

        if self.input_key is not None:
            output_key = self.input_key if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out



from pathlib import Path

import numpy as np
import jax.numpy as jnp

from flax.core import freeze, unfreeze

from ..utils.io import last_xyz_frame


from ..models import FENNIX

from ..utils.periodic_table import PERIODIC_TABLE_REV_IDX, ATOMIC_MASSES
from ..utils.atomic_units import AtomicUnits as au


def load_model(simulation_parameters):
    model_file = simulation_parameters.get("model_file")
    model_file = Path(str(model_file).strip())
    if not model_file.exists():
        raise FileNotFoundError(f"model file {model_file} not found")
    else:
        graph_config = simulation_parameters.get("graph_config", {})
        model = FENNIX.load(model_file, graph_config=graph_config)  # \
        print(f"# model_file: {model_file}")

    if "energy_terms" in simulation_parameters:
        energy_terms = simulation_parameters["energy_terms"]
        if isinstance(energy_terms, str):
            energy_terms = energy_terms.split()
        model.set_energy_terms(energy_terms)
        print("# energy terms:", model.energy_terms)

    return model


def load_system_data(simulation_parameters, fprec):
    ## LOAD SYSTEM CONFORMATION FROM FILES
    system_name = str(simulation_parameters.get("system", "system")).strip()
    indexed = simulation_parameters.get("xyz_input/indexed", True)
    has_comment_line = simulation_parameters.get("xyz_input/has_comment_line", False)
    xyzfile = Path(simulation_parameters.get("xyz_input/file", system_name + ".xyz"))
    if not xyzfile.exists():
        raise FileNotFoundError(f"xyz file {xyzfile} not found")
    system_name = str(simulation_parameters.get("system", xyzfile.stem)).strip()
    symbols, coordinates, _ = last_xyz_frame(
        xyzfile, indexed=indexed, has_comment_line=has_comment_line
    )
    coordinates = coordinates.astype(fprec)
    species = np.array([PERIODIC_TABLE_REV_IDX[s] for s in symbols], dtype=np.int32)
    nat = species.shape[0]

    ## GET MASS
    mass_amu = np.array(ATOMIC_MASSES, dtype=fprec)[species]
    deuterate = simulation_parameters.get("deuterate", False)
    if deuterate:
        print("# Replacing all hydrogens with deuteriums")
        mass_amu[species == 1] *= 2.0
    mass = mass_amu * (au.MPROT * (au.FS / au.BOHR) ** 2)

    ### GET TEMPERATURE
    temperature = np.clip(simulation_parameters.get("temperature", 300.0), 1.0e-6, None)
    kT = temperature / au.KELVIN
    totmass_amu = mass_amu.sum() / 6.02214129e-1

    ### GET TOTAL CHARGE
    total_charge = simulation_parameters.get("total_charge", 0.0)

    ### ENERGY UNIT
    energy_unit_str = simulation_parameters.get("energy_unit", "kcal/mol")
    energy_unit = au.get_multiplier(energy_unit_str)

    ## SYSTEM DATA
    system_data = {
        "name": system_name,
        "nat": nat,
        "symbols": symbols,
        "species": species,
        "mass": mass,
        "temperature": temperature,
        "kT": kT,
        "totmass_amu": totmass_amu,
        "total_charge": total_charge,
        "energy_unit": energy_unit,
        "energy_unit_str": energy_unit_str,
    }

    ### Set boundary conditions
    cell = simulation_parameters.get("cell", None)
    if cell is not None:
        cell = np.array(cell, dtype=fprec).reshape(3, 3)
        reciprocal_cell = np.linalg.inv(cell)
        volume = np.abs(np.linalg.det(cell))
        print("# cell matrix:")
        for l in cell:
            print("# ", l)
        # print(cell)
        dens = totmass_amu / volume
        print("# density: ", dens.item(), " g/cm^3")
        minimum_image = simulation_parameters.get("minimum_image", True)
        estimate_pressure = simulation_parameters.get("estimate_pressure", False)
        print("# minimum_image: ", minimum_image)

        crystal_input = simulation_parameters.get("xyz_input/crystal", False)
        if crystal_input:
            coordinates = coordinates @ cell

        pbc_data = {
            "cell": cell,
            "reciprocal_cell": reciprocal_cell,
            "volume": volume,
            "minimum_image": minimum_image,
            "estimate_pressure": estimate_pressure,
        }
    else:
        pbc_data = None
    system_data["pbc"] = pbc_data

    ### PIMD
    nbeads = simulation_parameters.get("nbeads", None)
    nreplicas = simulation_parameters.get("nreplicas", None)
    if nbeads is not None:
        nbeads = int(nbeads)
        print("# nbeads: ", nbeads)
        system_data["nbeads"] = nbeads
        coordinates = np.repeat(coordinates[None, :, :], nbeads, axis=0).reshape(-1, 3)
        species = np.repeat(species[None, :], nbeads, axis=0).reshape(-1)
        bead_index = np.arange(nbeads, dtype=np.int32).repeat(nat)
        natoms = np.array([nat] * nbeads, dtype=np.int32)

        eigmat = np.zeros((nbeads, nbeads))
        for i in range(nbeads - 1):
            eigmat[i, i] = 2.0
            eigmat[i, i + 1] = -1.0
            eigmat[i + 1, i] = -1.0
        eigmat[-1, -1] = 2.0
        eigmat[0, -1] = -1.0
        eigmat[-1, 0] = -1.0
        omk, eigmat = np.linalg.eigh(eigmat)
        omk[0] = 0.0
        omk = nbeads * kT * omk**0.5 / au.FS
        for i in range(nbeads):
            if eigmat[i, 0] < 0:
                eigmat[i] *= -1.0
        eigmat = jnp.asarray(eigmat, dtype=fprec)
        system_data["omk"] = omk
        system_data["eigmat"] = eigmat
        nreplicas = None
    elif nreplicas is not None:
        nreplicas = int(nreplicas)
        print("# nreplicas: ", nreplicas)
        system_data["nreplicas"] = nreplicas
        system_data["mass"] = np.repeat(mass[None, :], nreplicas, axis=0).reshape(-1)
        system_data["species"] = np.repeat(species[None, :], nreplicas, axis=0).reshape(
            -1
        )
        coordinates = np.repeat(coordinates[None, :, :], nreplicas, axis=0).reshape(
            -1, 3
        )
        species = np.repeat(species[None, :], nreplicas, axis=0).reshape(-1)
        bead_index = np.arange(nreplicas, dtype=np.int32).repeat(nat)
        natoms = np.array([nat] * nreplicas, dtype=np.int32)
    else:
        system_data["nreplicas"] = 1
        bead_index = np.array([0] * nat, dtype=np.int32)
        natoms = np.array([nat], dtype=np.int32)

    conformation = {
        "species": species,
        "coordinates": coordinates,
        "batch_index": bead_index,
        "natoms": natoms,
        "total_charge": total_charge,
    }
    if cell is not None:
        cell = cell[None, :, :]
        reciprocal_cell = reciprocal_cell[None, :, :]
        if nbeads is not None:
            cell = np.repeat(cell, nbeads, axis=0)
            reciprocal_cell = np.repeat(reciprocal_cell, nbeads, axis=0)
        elif nreplicas is not None:
            cell = np.repeat(cell, nreplicas, axis=0)
            reciprocal_cell = np.repeat(reciprocal_cell, nreplicas, axis=0)
        conformation["cells"] = cell
        conformation["reciprocal_cells"] = reciprocal_cell

    additional_keys = simulation_parameters.get("additional_keys", {})
    for key, value in additional_keys.items():
        conformation[key] = value

    return system_data, conformation


def initialize_preprocessing(simulation_parameters, model, conformation, system_data):
    nblist_verbose = simulation_parameters.get("nblist_verbose", False)
    nblist_skin = simulation_parameters.get("nblist_skin", -1.0)
    pbc_data = system_data.get("pbc", None)

    ### CONFIGURE PREPROCESSING
    preproc_state = unfreeze(model.preproc_state)
    layer_state = []
    for st in preproc_state["layers_state"]:
        stnew = unfreeze(st)
        if pbc_data is not None:
            stnew["minimum_image"] = pbc_data["minimum_image"]
        if nblist_skin > 0:
            stnew["nblist_skin"] = nblist_skin
        if "nblist_mult_size" in simulation_parameters:
            stnew["nblist_mult_size"] = simulation_parameters["nblist_mult_size"]
        if "nblist_add_neigh" in simulation_parameters:
            stnew["add_neigh"] = simulation_parameters["nblist_add_neigh"]
        layer_state.append(freeze(stnew))
    preproc_state["layers_state"] = layer_state
    preproc_state = freeze(preproc_state)

    ## initial preprocessing
    preproc_state = preproc_state.copy({"check_input": True})
    preproc_state, conformation = model.preprocessing(preproc_state, conformation)

    preproc_state = preproc_state.copy({"check_input": False})

    if nblist_verbose:
        graphs_keys = list(model._graphs_properties.keys())
        print("# graphs_keys: ", graphs_keys)
        print("# nblist state:", preproc_state)

    ### print model
    if simulation_parameters.get("print_model", False):
        print(model.summarize(example_data=conformation))

    return preproc_state, conformation






from typing import Any, Sequence, Callable, Union, Optional, Tuple, Dict
from copy import deepcopy
import dataclasses
from collections import OrderedDict

import jax
import jax.numpy as jnp
import flax.linen as nn
import numpy as np
from flax import serialization
from flax.core.frozen_dict import freeze, unfreeze, FrozenDict
from ..utils import AtomicUnits as au

from .preprocessing import (
    GraphGenerator,
    PreprocessingChain,
    JaxConverter,
    atom_unpadding,
    check_input,
)
from .modules import MODULES, PREPROCESSING, FENNIXModules


@dataclasses.dataclass
class FENNIX:
    """
    Static wrapper for FENNIX models

    The underlying model is a `fennol.models.modules.FENNIXModules` built from the `modules` dictionary
    which references registered modules in `fennol.models.modules.MODULES` and provides the parameters for initialization.

    Since the model is static and contains variables, it must be initialized right away with either
    `example_data`, `variables` or `rng_key`. If `variables` is provided, it is used directly. If `example_data`
    is provided, the model is initialized with `example_data` and the resulting variables are stored
    in the wrapper. If only `rng_key` is provided, the model is initialized with a dummy system and the resulting.
    """

    cutoff: Union[float, None]
    modules: FENNIXModules
    variables: Dict
    preprocessing: PreprocessingChain
    _apply: Callable[[Dict, Dict], Dict]
    _total_energy: Callable[[Dict, Dict], Tuple[jnp.ndarray, Dict]]
    _energy_and_forces: Callable[[Dict, Dict], Tuple[jnp.ndarray, jnp.ndarray, Dict]]
    _input_args: Dict
    _graphs_properties: Dict
    preproc_state: Dict
    energy_terms: Optional[Sequence[str]] = None
    _initializing: bool = True
    use_atom_padding: bool = False

    def __init__(
        self,
        cutoff: float,
        modules: OrderedDict,
        preprocessing: OrderedDict = OrderedDict(),
        example_data=None,
        rng_key: Optional[jax.random.PRNGKey] = None,
        variables: Optional[dict] = None,
        energy_terms: Optional[Sequence[str]] = None,
        use_atom_padding: bool = False,
        graph_config: Dict = {},
        energy_unit: str = "Ha",
        **kwargs,
    ) -> None:
        """Initialize the FENNIX model

        Arguments:
        ----------
        cutoff: float
            The cutoff radius for the model
        modules: OrderedDict
            The dictionary defining the sequence of FeNNol modules and their parameters.
        preprocessing: OrderedDict
            The dictionary defining the sequence of preprocessing modules and their parameters.
        example_data: dict
            Example data to initialize the model. If not provided, a dummy system is generated.
        rng_key: jax.random.PRNGKey
            The random key to initialize the model. If not provided, jax.random.PRNGKey(0) is used (should be avoided).
        variables: dict
            The variables of the model (i.e. weights, biases and all other tunable parameters).
            If not provided, the variables are initialized (usually at random)
        energy_terms: Sequence[str]
            The energy terms in the model output that will be summed to compute the total energy.
            If None, the total energy is always zero (useful for non-PES models).
        use_atom_padding: bool
            If True, the model will use atom padding for the input data.
            This is useful when one plans to frequently change the number of atoms in the system (for example during training).
        graph_config: dict
            Edit the graph configuration. Mostly used to change a long-range cutoff as a function of a simulation box size.

        """
        self._input_args = {
            "cutoff": cutoff,
            "modules": OrderedDict(modules),
            "preprocessing": OrderedDict(preprocessing),
            "energy_unit": energy_unit,
        }
        self.energy_unit = energy_unit
        self.Ha_to_model_energy = au.get_multiplier(energy_unit)
        self.cutoff = cutoff
        self.energy_terms = energy_terms
        self.use_atom_padding = use_atom_padding

        # add non-differentiable/non-jittable modules
        preprocessing = deepcopy(preprocessing)
        if cutoff is None:
            preprocessing_modules = []
        else:
            prep_keys = list(preprocessing.keys())
            graph_params = {"cutoff": cutoff, "graph_key": "graph"}
            if len(prep_keys) > 0 and prep_keys[0] == "graph":
                graph_params = {
                    **graph_params,
                    **preprocessing.pop("graph"),
                }
            graph_params = {**graph_params, **graph_config}

            preprocessing_modules = [
                GraphGenerator(**graph_params),
            ]

        for name, params in preprocessing.items():
            key = str(params.pop("module_name")) if "module_name" in params else name
            key = str(params.pop("FID")) if "FID" in params else key
            mod = PREPROCESSING[key.upper()](**freeze(params))
            preprocessing_modules.append(mod)

        self.preprocessing = PreprocessingChain(
            tuple(preprocessing_modules), use_atom_padding
        )
        graphs_properties = self.preprocessing.get_graphs_properties()
        self._graphs_properties = freeze(graphs_properties)
        # add preprocessing modules that should be differentiated/jitted
        mods = [(JaxConverter, {})] + self.preprocessing.get_processors()
        # mods = self.preprocessing.get_processors(return_list=True)

        # build the model
        modules = deepcopy(modules)
        modules_names = []
        for name, params in modules.items():
            key = str(params.pop("module_name")) if "module_name" in params else name
            key = str(params.pop("FID")) if "FID" in params else key
            if name in modules_names:
                raise ValueError(f"Module {name} already exists")
            modules_names.append(name)
            params["name"] = name
            mod = MODULES[key.upper()]
            fields = [f.name for f in dataclasses.fields(mod)]
            if "_graphs_properties" in fields:
                params["_graphs_properties"] = graphs_properties
            if "_energy_unit" in fields:
                params["_energy_unit"] = energy_unit
            mods.append((mod, params))

        self.modules = FENNIXModules(mods)

        self.__apply = self.modules.apply
        self._apply = jax.jit(self.modules.apply)

        self.set_energy_terms(energy_terms)

        # initialize the model

        inputs, rng_key = self.reinitialize_preprocessing(rng_key, example_data)

        if variables is not None:
            self.variables = variables
        elif rng_key is not None:
            self.variables = self.modules.init(rng_key, inputs)
        else:
            raise ValueError(
                "Either variables or a jax.random.PRNGKey must be provided for initialization"
            )

        self._initializing = False

    def set_energy_terms(
        self, energy_terms: Union[Sequence[str], None], jit: bool = True
    ) -> None:
        """Set the energy terms to be computed by the model and prepare the energy and force functions."""
        object.__setattr__(self, "energy_terms", energy_terms)
        if isinstance(energy_terms, str):
            energy_terms = [energy_terms]
        
        if energy_terms is None or len(energy_terms) == 0:

            def total_energy(variables, data):
                out = self.__apply(variables, data)
                coords = out["coordinates"]
                nsys = out["natoms"].shape[0]
                nat = coords.shape[0]
                dtype = coords.dtype
                e = jnp.zeros(nsys, dtype=dtype)
                eat = jnp.zeros(nat, dtype=dtype)
                out["total_energy"] = e
                out["atomic_energies"] = eat
                return e, out

            def energy_and_forces(variables, data):
                e, out = total_energy(variables, data)
                f = jnp.zeros_like(out["coordinates"])
                out["forces"] = f
                return e, f, out

            def energy_and_forces_and_virial(variables, data):
                e, f, out = energy_and_forces(variables, data)
                v = jnp.zeros(
                    (out["natoms"].shape[0], 3, 3), dtype=out["coordinates"].dtype
                )
                out["virial_tensor"] = v
                return e, f, v, out

        else:
            # build the energy and force functions
            def total_energy(variables, data):
                out = self.__apply(variables, data)
                atomic_energies = 0.0
                system_energies = 0.0
                species = out["species"]
                nsys = out["natoms"].shape[0]
                for term in self.energy_terms:
                    e = out[term]
                    if e.ndim > 1 and e.shape[-1] == 1:
                        e = jnp.squeeze(e, axis=-1)
                    if e.shape[0] == nsys and nsys != species.shape[0]:
                        system_energies += e
                        continue
                    assert e.shape == species.shape
                    atomic_energies += e
                # atomic_energies = jnp.squeeze(atomic_energies, axis=-1)
                if isinstance(atomic_energies, jnp.ndarray):
                    if "true_atoms" in out:
                        atomic_energies = jnp.where(
                            out["true_atoms"], atomic_energies, 0.0
                        )
                    out["atomic_energies"] = atomic_energies
                    energies = jax.ops.segment_sum(
                        atomic_energies,
                        data["batch_index"],
                        num_segments=len(data["natoms"]),
                    )
                else:
                    energies = 0.0

                if isinstance(system_energies, jnp.ndarray):
                    if "true_sys" in out:
                        system_energies = jnp.where(
                            out["true_sys"], system_energies, 0.0
                        )
                    out["system_energies"] = system_energies

                out["total_energy"] = energies + system_energies
                return out["total_energy"], out

            def energy_and_forces(variables, data):
                def _etot(variables, coordinates):
                    energy, out = total_energy(
                        variables, {**data, "coordinates": coordinates}
                    )
                    return energy.sum(), out

                de, out = jax.grad(_etot, argnums=1, has_aux=True)(
                    variables, data["coordinates"]
                )
                out["forces"] = -de

                return out["total_energy"], out["forces"], out

            # def energy_and_forces_and_virial(variables, data):
            #     x = data["coordinates"]
            #     batch_index = data["batch_index"]
            #     if "cells" in data:
            #         cells = data["cells"]
            #         ## cells is a nbatchx3x3 matrix which lines are cell vectors (i.e. cells[0,0,:] is the first cell vector of the first system)

            #         def _etot(variables, coordinates, cells):
            #             reciprocal_cells = jnp.linalg.inv(cells)
            #             energy, out = total_energy(
            #                 variables,
            #                 {
            #                     **data,
            #                     "coordinates": coordinates,
            #                     "cells": cells,
            #                     "reciprocal_cells": reciprocal_cells,
            #                 },
            #             )
            #             return energy.sum(), out

            #         (dedx, dedcells), out = jax.grad(_etot, argnums=(1, 2), has_aux=True)(
            #             variables, x, cells
            #         )
            #         f= -dedx
            #         out["forces"] = f
            #     else:
            #         _,f,out = energy_and_forces(variables, data)

            #     vir = -jax.ops.segment_sum(
            #         f[:, :, None] * x[:, None, :],
            #         batch_index,
            #         num_segments=len(data["natoms"]),
            #     )

            #     if "cells" in data:
            #         # dvir = jax.vmap(jnp.matmul)(dedcells, cells.transpose(0, 2, 1))
            #         dvir = jnp.einsum("...ki,...kj->...ij", dedcells, cells)
            #         nsys = data["natoms"].shape[0]
            #         if cells.shape[0]==1 and nsys>1:
            #             dvir = dvir / nsys
            #         vir = vir + dvir

            #     out["virial_tensor"] = vir

            #     return out["total_energy"], f, vir, out

            def energy_and_forces_and_virial(variables, data):
                x = data["coordinates"]
                scaling = jnp.asarray(
                    np.eye(3)[None, :, :].repeat(data["natoms"].shape[0], axis=0)
                )
                def _etot(variables, coordinates, scaling):
                    batch_index = data["batch_index"]
                    coordinates = jax.vmap(jnp.matmul)(
                        coordinates, scaling[batch_index]
                    )
                    inputs = {**data, "coordinates": coordinates}
                    if "cells" in data:
                        ## cells is a nbatchx3x3 matrix which lines are cell vectors (i.e. cells[0,0,:] is the first cell vector of the first system)
                        cells = jax.vmap(jnp.matmul)(data["cells"], scaling)
                        reciprocal_cells = jnp.linalg.inv(cells)
                        inputs["cells"] = cells
                        inputs["reciprocal_cells"] = reciprocal_cells
                    energy, out = total_energy(variables, inputs)
                    return energy.sum(), out

                (dedx, vir), out = jax.grad(_etot, argnums=(1, 2), has_aux=True)(
                    variables, x, scaling
                )
                f = -dedx
                out["forces"] = f
                out["virial_tensor"] = vir

                return out["total_energy"], f, vir, out

        object.__setattr__(self, "_total_energy_raw", total_energy)
        if jit:
            object.__setattr__(self, "_total_energy", jax.jit(total_energy))
            object.__setattr__(self, "_energy_and_forces", jax.jit(energy_and_forces))
            object.__setattr__(
                self,
                "_energy_and_forces_and_virial",
                jax.jit(energy_and_forces_and_virial),
            )
        else:
            object.__setattr__(self, "_total_energy", total_energy)
            object.__setattr__(self, "_energy_and_forces", energy_and_forces)
            object.__setattr__(
                self, "_energy_and_forces_and_virial", energy_and_forces_and_virial
            )

    def get_gradient_function(
        self,
        *gradient_keys: Sequence[str],
        jit: bool = True,
        variables_as_input: bool = False,
    ):
        """Return a function that computes the energy and the gradient of the energy with respect to the keys in gradient_keys"""

        def _energy_gradient(variables, data):
            def _etot(variables, inputs):
                if "strain" in inputs:
                    scaling = inputs["strain"]
                    batch_index = data["batch_index"]
                    coordinates = inputs["coordinates"] if "coordinates" in inputs else data["coordinates"]
                    coordinates = jax.vmap(jnp.matmul)(
                        coordinates, scaling[batch_index]
                    )
                    inputs = {**inputs, "coordinates": coordinates}
                    if "cells" in inputs or "cells" in data:
                        cells = inputs["cells"] if "cells" in inputs else data["cells"]
                        cells = jax.vmap(jnp.matmul)(cells, scaling)
                        inputs["cells"] = cells
                if "cells" in inputs:
                    reciprocal_cells = jnp.linalg.inv(inputs["cells"])
                    inputs = {**inputs, "reciprocal_cells": reciprocal_cells}
                energy, out = self._total_energy_raw(variables, {**data, **inputs})
                return energy.sum(), out

            if "strain" in gradient_keys and "strain" not in data:
                data = {**data, "strain": jnp.array(np.eye(3)[None, :, :].repeat(data["natoms"].shape[0], axis=0))}
            inputs = {k: data[k] for k in gradient_keys}
            de, out = jax.grad(_etot, argnums=1, has_aux=True)(variables, inputs)

            return (
                out["total_energy"],
                de,
                {**out, **{"dEd_" + k: de[k] for k in gradient_keys}},
            )

        if variables_as_input:
            energy_gradient = _energy_gradient
        else:

            def energy_gradient(data):
                return _energy_gradient(self.variables, data)

        if jit:
            return jax.jit(energy_gradient)
        else:
            return energy_gradient

    def preprocess(self,use_gpu=False, verbose=False,**inputs) -> Dict[str, Any]:
        """apply preprocessing to the input data

        !!! This is not a pure function => do not apply jax transforms !!!"""
        if self.preproc_state is None:
            out, _ = self.reinitialize_preprocessing(example_data=inputs)
        elif use_gpu:
            do_check_input = self.preproc_state.get("check_input", True)
            if do_check_input:
                inputs = check_input(inputs)
            preproc_state, inputs = self.preprocessing.atom_padding(
                self.preproc_state, inputs
            )
            inputs = self.preprocessing.process(preproc_state, inputs)
            preproc_state, state_up, out, overflow = (
                self.preprocessing.check_reallocate(
                    preproc_state, inputs
                )
            )
            if verbose and overflow:
                print("GPU preprocessing: nblist overflow => reallocating nblist")
                print("size updates:", state_up)
        else:
            preproc_state, out = self.preprocessing(self.preproc_state, inputs)

        object.__setattr__(self, "preproc_state", preproc_state)
        return out

    def reinitialize_preprocessing(
        self, rng_key: Optional[jax.random.PRNGKey] = None, example_data=None
    ) -> None:
        ### TODO ###
        if rng_key is None:
            rng_key_pre = jax.random.PRNGKey(0)
        else:
            rng_key, rng_key_pre = jax.random.split(rng_key)

        if example_data is None:
            rng_key_sys, rng_key_pre = jax.random.split(rng_key_pre)
            example_data = self.generate_dummy_system(rng_key_sys, n_atoms=10)

        preproc_state, inputs = self.preprocessing.init_with_output(example_data)
        object.__setattr__(self, "preproc_state", preproc_state)
        return inputs, rng_key

    def __call__(self, variables: Optional[dict] = None, gpu_preprocessing=False,**inputs) -> Dict[str, Any]:
        """Apply the FENNIX model (preprocess + modules)

        !!! This is not a pure function => do not apply jax transforms !!!
        if you want to apply jax transforms, use  self._apply(variables, inputs) which is pure and preprocess the input using self.preprocess
        """
        if variables is None:
            variables = self.variables
        inputs = self.preprocess(use_gpu=gpu_preprocessing,**inputs)
        output = self._apply(variables, inputs)
        if self.use_atom_padding:
            output = atom_unpadding(output)
        return output

    def total_energy(
        self, variables: Optional[dict] = None, unit: Union[float, str] = None, gpu_preprocessing=False,**inputs
    ) -> Tuple[jnp.ndarray, Dict]:
        """compute the total energy of the system

        !!! This is not a pure function => do not apply jax transforms !!!
        if you want to apply jax transforms, use self._total_energy(variables, inputs) which is pure and preprocess the input using self.preprocess
        """
        if variables is None:
            variables = self.variables
        inputs = self.preprocess(use_gpu=gpu_preprocessing,**inputs)
        # def print_shape(path,value):
        #     if isinstance(value,jnp.ndarray):
        #         print(path,value.shape)
        #     else:
        #         print(path,value)
        # jax.tree_util.tree_map_with_path(print_shape,inputs)
        _, output = self._total_energy(variables, inputs)
        if self.use_atom_padding:
            output = atom_unpadding(output)
        e = output["total_energy"]
        if unit is not None:
            model_energy_unit = self.Ha_to_model_energy
            if isinstance(unit, str):
                unit = au.get_multiplier(unit)
            e = e * (unit / model_energy_unit)
        return e, output

    def energy_and_forces(
        self, variables: Optional[dict] = None, unit: Union[float, str] = None,gpu_preprocessing=False,**inputs
    ) -> Tuple[jnp.ndarray, jnp.ndarray, Dict]:
        """compute the total energy and forces of the system

        !!! This is not a pure function => do not apply jax transforms !!!
        if you want to apply jax transforms, use  self._energy_and_forces(variables, inputs) which is pure and preprocess the input using self.preprocess
        """
        if variables is None:
            variables = self.variables
        inputs = self.preprocess(use_gpu=gpu_preprocessing,**inputs)
        _, _, output = self._energy_and_forces(variables, inputs)
        if self.use_atom_padding:
            output = atom_unpadding(output)
        e = output["total_energy"]
        f = output["forces"]
        if unit is not None:
            model_energy_unit = self.Ha_to_model_energy
            if isinstance(unit, str):
                unit = au.get_multiplier(unit)
            e = e * (unit / model_energy_unit)
            f = f * (unit / model_energy_unit)
        return e, f, output

    def energy_and_forces_and_virial(
        self, variables: Optional[dict] = None, unit: Union[float, str] = None,gpu_preprocessing=False, **inputs
    ) -> Tuple[jnp.ndarray, jnp.ndarray, Dict]:
        """compute the total energy and forces of the system

        !!! This is not a pure function => do not apply jax transforms !!!
        if you want to apply jax transforms, use  self._energy_and_forces_and_virial(variables, inputs) which is pure and preprocess the input using self.preprocess
        """
        if variables is None:
            variables = self.variables
        inputs = self.preprocess(use_gpu=gpu_preprocessing,**inputs)
        _, _, _, output = self._energy_and_forces_and_virial(variables, inputs)
        if self.use_atom_padding:
            output = atom_unpadding(output)
        e = output["total_energy"]
        f = output["forces"]
        v = output["virial_tensor"]
        if unit is not None:
            model_energy_unit = self.Ha_to_model_energy
            if isinstance(unit, str):
                unit = au.get_multiplier(unit)
            e = e * (unit / model_energy_unit)
            f = f * (unit / model_energy_unit)
            v = v * (unit / model_energy_unit)
        return e, f, v, output

    def remove_atom_padding(self, output):
        """remove atom padding from the output"""
        return atom_unpadding(output)

    def get_model(self) -> Tuple[FENNIXModules, Dict]:
        """return the model and its variables"""
        return self.modules, self.variables

    def get_preprocessing(self) -> Tuple[PreprocessingChain, Dict]:
        """return the preprocessing chain and its state"""
        return self.preprocessing, self.preproc_state

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "variables":
            if __value is not None:
                if not (
                    isinstance(__value, dict)
                    or isinstance(__value, OrderedDict)
                    or isinstance(__value, FrozenDict)
                ):
                    raise ValueError(f"{__name} must be a dict")
                object.__setattr__(self, __name, JaxConverter()(__value))
            else:
                raise ValueError(f"{__name} cannot be None")
        elif __name == "preproc_state":
            if __value is not None:
                if not (
                    isinstance(__value, dict)
                    or isinstance(__value, OrderedDict)
                    or isinstance(__value, FrozenDict)
                ):
                    raise ValueError(f"{__name} must be a FrozenDict")
                object.__setattr__(self, __name, freeze(JaxConverter()(__value)))
            else:
                raise ValueError(f"{__name} cannot be None")

        elif self._initializing:
            object.__setattr__(self, __name, __value)
        else:
            raise ValueError(f"{__name} attribute of FENNIX model is immutable.")

    def generate_dummy_system(
        self, rng_key: jax.random.PRNGKey, box_size=None, n_atoms: int = 10
    ) -> Dict[str, Any]:
        """
        Generate dummy system for initialization
        """
        if box_size is None:
            box_size = 2 * self.cutoff
            for g in self._graphs_properties.values():
                cutoff = g["cutoff"]
                if cutoff is not None:
                    box_size = min(box_size, 2 * g["cutoff"])
        coordinates = np.array(
            jax.random.uniform(rng_key, (n_atoms, 3), maxval=box_size), dtype=np.float64
        )
        species = np.ones((n_atoms,), dtype=np.int32)
        batch_index = np.zeros((n_atoms,), dtype=np.int32)
        natoms = np.array([n_atoms], dtype=np.int32)
        return {
            "species": species,
            "coordinates": coordinates,
            # "graph": graph,
            "batch_index": batch_index,
            "natoms": natoms,
        }

    def summarize(
        self, rng_key: jax.random.PRNGKey = None, example_data=None, **kwargs
    ) -> str:
        """Summarize the model architecture and parameters"""
        if rng_key is None:
            head = "Summarizing with example data:\n"
            rng_key = jax.random.PRNGKey(0)
        if example_data is None:
            head = "Summarizing with dummy 10 atoms system:\n"
            rng_key, rng_key_sys = jax.random.split(rng_key)
            example_data = self.generate_dummy_system(rng_key_sys, n_atoms=10)
        rng_key, rng_key_pre = jax.random.split(rng_key)
        _, inputs = self.preprocessing.init_with_output(example_data)
        return head + nn.tabulate(self.modules, rng_key, **kwargs)(inputs)

    def to_dict(self):
        """return a dictionary representation of the model"""
        return {
            **self._input_args,
            "energy_terms": self.energy_terms,
            "variables": deepcopy(self.variables),
        }

    def save(self, filename):
        """save the model to a file"""
        state_dict = self.to_dict()
        state_dict["preprocessing"] = [
            [k, v] for k, v in state_dict["preprocessing"].items()
        ]
        state_dict["modules"] = [[k, v] for k, v in state_dict["modules"].items()]
        with open(filename, "wb") as f:
            f.write(serialization.msgpack_serialize(state_dict))

    @classmethod
    def load(
        cls,
        filename,
        use_atom_padding=False,
        graph_config={},
    ):
        """load a model from a file"""
        with open(filename, "rb") as f:
            state_dict = serialization.msgpack_restore(f.read())
        state_dict["preprocessing"] = {k: v for k, v in state_dict["preprocessing"]}
        state_dict["modules"] = {k: v for k, v in state_dict["modules"]}
        return cls(
            **state_dict,
            graph_config=graph_config,
            use_atom_padding=use_atom_padding,
        )



import flax.linen as nn
from typing import Any, Sequence, Callable, Union, ClassVar,Optional
import jax.numpy as jnp
import jax
import numpy as np
from functools import partial
import dataclasses
from ...utils.periodic_table import PERIODIC_TABLE


class EnsembleStatistics(nn.Module):
    """Computes the mean and variance of an ensemble.
    
    FID: ENSEMBLE_STAT
    """

    key: str
    """The key to access the input data from the `inputs` dictionary."""
    axis: int = -1
    """The axis along which to compute the mean and variance."""
    shuffle_ensemble: bool = False
    """Whether to shuffle the ensemble."""
    weights_key: Optional[str] = None
    mean_key: Optional[str] = None

    FID: ClassVar[str] = "ENSEMBLE_STAT"

    @nn.compact
    def __call__(self, inputs) -> Any:
        x = inputs[self.key]
        if self.weights_key is not None:
            weights = inputs[self.weights_key]
        else:
            weights = jnp.ones_like(x)
        weights = weights/jnp.sum(weights,axis=self.axis,keepdims=True)
        mean = jnp.sum(x*weights, axis=self.axis,keepdims=True)

        nsamples = x.shape[self.axis]
        if nsamples == 1:
            var = jnp.zeros_like(mean)
        else:
            # var = jnp.var(x, axis=self.axis, ddof=1)
            var = jnp.sum(weights*(x-mean)**2,axis=self.axis)*(nsamples/(nsamples-1.))
        
        mean = jnp.squeeze(mean,axis=self.axis)
        output = {**inputs, self.key + "_mean": mean, self.key + "_var": var}

        if self.mean_key is not None:
            output[self.mean_key] = mean

        training = "training" in inputs.get("flags", {})
        if self.shuffle_ensemble and training and "rng_key" in inputs:
            key, subkey = jax.random.split(inputs["rng_key"])
            x = jax.random.permutation(subkey, x, axis=self.axis, independent=True)
            output[self.key] = x
            output["rng_key"] = key
        return output


class EnsembleShift(nn.Module):
    """Shifts the mean of an ensemble to match a reference tensor.
    
    FID: ENSEMBLE_SHIFT
    """

    key: str
    """The key to access the input data from the `inputs` dictionary."""
    ref_key: str
    """The key to access the reference data from the `inputs` dictionary."""
    axis: int = -1
    """The axis of the ensemble."""
    output_key: Optional[str] = None
    """The key of the output ensemble. If None, the input key will be used."""


    FID: ClassVar[str] = "ENSEMBLE_SHIFT"

    @nn.compact
    def __call__(self, inputs) -> Any:
        x = inputs[self.key]
        mean = jnp.mean(x, axis=self.axis, keepdims=True)
        ref = inputs[self.ref_key].reshape(mean.shape)
        x = x - mean + ref
        output_key = self.key if self.output_key is None else self.output_key
        return {**inputs, output_key: x}


class ConstrainEvidence(nn.Module):
    """ Constrain the parameters of an evidential model

    FID: CONSTRAIN_EVIDENCE

    ### References
    - Amini et al, Deep Evidential Regression, NeurIPS 2020 (https://arxiv.org/abs/1910.02600)
    - Meinert et al, Multivariate Deep Evidential Regression, (https://arxiv.org/pdf/2104.06135.pdf)

    """
    key: str
    """The key to access the input data from the `inputs` dictionary."""
    output_key: Optional[str] = None
    """The key to use for the constrained paramters in the `output` dictionary."""
    beta_scale: Union[str, float] = 1.0
    """The scale of the beta parameter."""
    alpha_init: float = 2.0
    """The initial value of the alpha parameter."""
    nu_init: float = 1.0
    """The initial value of the nu parameter."""
    chemical_shift: Optional[float] = None
    """The initial chemical shift of evidence."""
    trainable_beta: bool = False
    """Whether the beta parameter is trainable."""
    constant_beta: bool = False
    """Whether the beta parameter is a constant."""
    trainable_alpha: bool = False
    """Whether the alpha parameter is trainable."""
    constant_alpha: bool = False
    """Whether the alpha parameter is a constant."""
    trainable_nu: bool = False
    """Whether the nu parameter is trainable."""
    constant_nu: bool = False
    """Whether the nu parameter is a constant."""
    nualpha_coupling: Optional[float] = None
    """The coupling constant between nu and alpha."""
    graph_key: Optional[str] = None
    """The key to access the graph data from the `inputs` dictionary. 
        Only used to obtain an environment-dependent chemical shift."""
    self_weight: float = 10.0
    """The weight of the self interaction in environment-dependent chemical shift."""
    # target_dim: Optional[int] = None

    FID: ClassVar[str] = "CONSTRAIN_EVIDENCE"

    @nn.compact
    def __call__(self, inputs) -> Any:
        x = inputs[self.key]
        dim = 3
        assert x.shape[-1] == dim, f"Dimension {-1} must be {dim}, got {x.shape[-1]}"
        nui, alphai, betai = jnp.split(x, 3, axis=-1)

        if self.chemical_shift is not None:
            # modify nu evidence depending on the chemical species 
            #   this is useful for predicting high uncertainties for unknown species
            #   in known geometries
            nu_shift = jnp.abs(
                self.param(
                    "nu_shift",
                    lambda key, shape: jnp.ones(shape) * self.chemical_shift,
                    (len(PERIODIC_TABLE),),
                )
            )[inputs["species"]]
            if self.graph_key is not None:
                graph = inputs[self.graph_key]
                edge_dst = graph["edge_dst"]
                edge_src = graph["edge_src"]
                switch = graph["switch"]
                nushift_neigh = jax.ops.segment_sum(
                    nu_shift[edge_dst] * switch, edge_src, nu_shift.shape[0]
                )
                norm = self.self_weight + jax.ops.segment_sum(
                    switch, edge_src, nu_shift.shape[0]
                )
                nu_shift = (self.self_weight * nu_shift + nushift_neigh) / norm
            nu_shift = nu_shift[:, None]
        else:
            nu_shift = 1.0

        if self.nualpha_coupling is None:
            # couple nu and alpha to remove overparameterization
            #  of the evidential model (see Meinert et al)
            if self.constant_alpha:
                if self.trainable_alpha:
                    alphai = 1 + jnp.abs(
                        self.param(
                            "alpha",
                            lambda key: jnp.asarray(self.alpha_init - 1),
                        )
                    ) * jnp.ones_like(alphai)
                else:
                    assert self.alpha_init > 1, "alpha_init must be >1"
                    alphai = self.alpha_init * jnp.ones_like(alphai)
            elif self.constant_nu:
                alphai = 1 + (self.alpha_init - 1) * nu_shift * jax.nn.softplus(alphai)
            else:
                alphai = 1 + (self.alpha_init - 1) * jax.nn.softplus(alphai)

            if self.constant_nu:
                if self.trainable_nu:
                    nui = 1.0e-5 + jnp.abs(
                        self.param(
                            "nu",
                            lambda key: jnp.asarray(self.nu_init),
                        )
                    ) * jnp.ones_like(nui)
                else:
                    nui = self.nu_init * jnp.ones_like(nui)
            else:
                nui = 1.0e-5 + self.nu_init * nu_shift * jax.nn.softplus(nui)
        else:
            alphai = 1 + nu_shift * jax.nn.softplus(alphai)
            if self.trainable_nu:
                nualpha_coupling = jnp.abs(
                    self.param(
                        "nualpha_coupling",
                        lambda key: jnp.asarray(self.nualpha_coupling),
                    )
                )
            else:
                nualpha_coupling = self.nualpha_coupling
            nui = nualpha_coupling * 2 * alphai

        if self.constant_beta:
            if self.trainable_beta:
                betai = (
                    jax.nn.softplus(
                        self.param(
                            "beta",
                            lambda key: jnp.asarray(0.0),
                        )
                    )
                    / np.log(2.0)
                ) * jnp.ones_like(alphai)
            else:
                betai = jnp.ones_like(alphai)
        else:
            betai = jax.nn.softplus(betai)

        betai = self.beta_scale * betai

        output = jnp.concatenate([nui, alphai, betai], axis=-1)
        # if self.target_dim is not None:
        #     output = output[...,None,:].repeat(self.target_dim, axis=-2)

        output_key = self.key if self.output_key is None else self.output_key
        out = {
            **inputs,
            output_key: output,
            output_key + "_var": jnp.squeeze(betai / (nui * (alphai - 1)), axis=-1),
            output_key + "_aleatoric": jnp.squeeze(betai / (alphai - 1), axis=-1),
            output_key
            + "_wst2": jnp.squeeze(betai * (1 + nui) / (alphai * nui), axis=-1),
        }
        return out



import ase
import ase.calculators.calculator
import ase.units
import numpy as np
import jax.numpy as jnp
from . import FENNIX
from .models.preprocessing import convert_to_jax
from typing import Sequence, Union, Optional
from ase.stress import full_3x3_to_voigt_6_stress
import jax
from .utils import AtomicUnits as au

class FENNIXCalculator(ase.calculators.calculator.Calculator):
    """FENNIX calculator for ASE.

    Arguments:
    ----------
    model: str or FENNIX
        The path to the model or the model itself.
    use_atom_padding: bool, default=False
        Whether to use atom padding or not. Atom padding is useful to prevent recompiling the model if the number of atoms changes. If the number of atoms is expected to be fixed, it is recommended to set this to False.
    gpu_preprocessing: bool, default=False
        Whether to preprocess the data on the GPU or not. This is useful for large systems, but may not be necessary (or even slower) for small systems.
    atoms: ase.Atoms, default=None
        The atoms object to be used for the calculation. If provided, the calculator will be initialized with the atoms object.
    verbose: bool, default=False
        Whether to print nblist update information or not.
    energy_terms: list of str, default=None
        The energy terms to include in the total energy. If None, this will default to the energy terms defined in the model.
    """

    implemented_properties = ["energy", "forces", "stress"]

    def __init__(
        self,
        model: Union[str, FENNIX],
        gpu_preprocessing: bool = False,
        atoms: Optional[ase.Atoms] = None,
        verbose: bool = False,
        energy_terms: Optional[Sequence[str]] = None,
        use_float64: bool = False,
        matmul_prec: Optional[str] = None,
        save_raw_output: bool = False,
        **kwargs
    ):
        super().__init__()
        if use_float64:
            jax.config.update("jax_enable_x64", True)
        if matmul_prec is not None:
            assert matmul_prec in [
                "default",
                "high",
                "highest",
            ], "matmul_prec should be one of 'default', 'high', 'highest'"
            jax.config.update("jax_default_matmul_precision", matmul_prec)

        if isinstance(model, FENNIX):
            self.model = model
        else:
            self.model = FENNIX.load(model, **kwargs)
        if energy_terms is not None:
            self.model.set_energy_terms(energy_terms)
        self.dtype = "float64" if use_float64 else "float32"
        self.gpu_preprocessing = gpu_preprocessing
        self.verbose = verbose
        self._fennol_inputs = None
        self._raw_inputs = None
        self.save_raw_output = save_raw_output

        model_unit = au.get_multiplier(self.model.energy_unit)
        self.energy_conv = ase.units.Hartree / model_unit
        if atoms is not None:
            self.preprocess(atoms)

    def calculate(
        self,
        atoms=None,
        properties=["energy"],
        system_changes=ase.calculators.calculator.all_changes,
    ):
        super().calculate(atoms, properties, system_changes)
        inputs = self.preprocess(self.atoms, system_changes=system_changes)
        total_charge = self.atoms.get_initial_charges().sum()
        inputs["total_charge"] = int(total_charge)

        results = {}
        if "stress" in properties:
            e, f, virial, output = self.model._energy_and_forces_and_virial(
                self.model.variables, inputs
            )
            volume = self.atoms.get_volume()
            stress = np.asarray(virial[0]) * self.energy_conv / volume
            results["stress"] = full_3x3_to_voigt_6_stress(stress)
            results["forces"] = np.asarray(f) * self.energy_conv
        elif "forces" in properties:
            e, f, output = self.model._energy_and_forces(self.model.variables, inputs)
            results["forces"] = np.asarray(f) * self.energy_conv
        else:
            e, output = self.model._total_energy(self.model.variables, inputs)

        results["energy"] = float(e[0]) * self.energy_conv
        if self.model.use_atom_padding and "forces" in results:
            mask = np.asarray(output["true_atoms"])
            results["forces"] = results["forces"][mask]
        
        if self.save_raw_output:
            results["raw_output"] = output

        self.results.update(results)

    def preprocess(self, atoms, system_changes=ase.calculators.calculator.all_changes):

        force_cpu_preprocessing = False
        if self._raw_inputs is None:
            force_cpu_preprocessing = True
            cell = np.asarray(atoms.get_cell(complete=True).array, dtype=self.dtype)
            pbc = np.asarray(atoms.get_pbc(), dtype=bool)
            if np.all(pbc):
                use_pbc = True
            elif np.any(pbc):
                raise NotImplementedError("PBC should be activated in all directions.")
            else:
                use_pbc = False

            species = np.asarray(atoms.get_atomic_numbers(), dtype=np.int32)
            coordinates = np.asarray(atoms.get_positions(), dtype=self.dtype)
            natoms = np.array([len(species)], dtype=np.int32)
            batch_index = np.array([0] * len(species), dtype=np.int32)

            inputs = {
                "species": species,
                "coordinates": coordinates,
                "natoms": natoms,
                "batch_index": batch_index,
            }
            if use_pbc:
                reciprocal_cell = np.linalg.inv(cell)
                inputs["cells"] = cell.reshape(1, 3, 3)
                inputs["reciprocal_cells"] = reciprocal_cell.reshape(1, 3, 3)
            self._raw_inputs = convert_to_jax(inputs)
        else:
            if "cell" in system_changes:
                pbc = np.asarray(atoms.get_pbc(), dtype=bool)
                if np.all(pbc):
                    use_pbc = True
                elif np.any(pbc):
                    raise NotImplementedError(
                        "PBC should be activated in all directions."
                    )
                else:
                    use_pbc = False
                if use_pbc:
                    cell = np.asarray(
                        atoms.get_cell(complete=True).array, dtype=self.dtype
                    )
                    reciprocal_cell = np.linalg.inv(cell)
                    self._raw_inputs["cells"] = jnp.asarray(cell.reshape(1, 3, 3))
                    self._raw_inputs["reciprocal_cells"] = jnp.asarray(
                        reciprocal_cell.reshape(1, 3, 3)
                    )
                elif "cells" in self._raw_inputs:
                    del self._raw_inputs["cells"]
                    del self._raw_inputs["reciprocal_cells"]
            if "numbers" in system_changes:
                self._raw_inputs["species"] = jnp.asarray(
                    atoms.get_atomic_numbers(), dtype=jnp.int32
                )
                self._raw_inputs["natoms"] = jnp.array(
                    [len(self._raw_inputs["species"])], dtype=np.int32
                )
                self._raw_inputs["batch_index"] = jnp.array(
                    [0] * len(self._raw_inputs["species"]), dtype=np.int32
                )
                force_cpu_preprocessing = True
            if "positions" in system_changes:
                self._raw_inputs["coordinates"] = jnp.asarray(
                    atoms.get_positions(), dtype=self.dtype
                )

        if self.gpu_preprocessing and not force_cpu_preprocessing:
            _, inputs = self.model.preprocessing.atom_padding(
                self.model.preproc_state, self._raw_inputs
            )
            inputs = {**self._fennol_inputs, **inputs}

            inputs = self.model.preprocessing.process(self.model.preproc_state, inputs)
            self.model.preproc_state, state_up, inputs, overflow = (
                self.model.preprocessing.check_reallocate(
                    self.model.preproc_state, inputs
                )
            )
            self._fennol_inputs = inputs
            if self.verbose and overflow:
                print("FENNIX nblist overflow => reallocating nblist")
                print("  size updates:", state_up)
        else:
            self._fennol_inputs = self.model.preprocess(**self._raw_inputs)

        return self._fennol_inputs



import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Sequence, Dict, Union, ClassVar, Optional
import numpy as np
from ...utils.periodic_table import (
    VALENCE_ELECTRONS,
)

class ChargeHypothesis(nn.Module):
    """Embedding with total charge constraint vi Multiple Neural Charge Equilibration.

    FID: CHARGE_HYPOTHESIS

    """

    embedding_key: str
    """Key for the embedding in the inputs that is used to predict an 'electron affinity' weight"""
    output_key: Union[str, None] = None
    """Key for the charges in the outputs"""
    total_charge_key: str = "total_charge"
    """Key for the total charge in the inputs"""
    ncharges: int = 10
    """The number of charge hypothesis"""
    mode: str = "qeq"
    """Charge distribution mode. Only 'qeq' available for now."""
    squeeze: bool = True

    FID: ClassVar[str] = "CHARGE_HYPOTHESIS"

    @nn.compact
    def __call__(self, inputs):
        embedding = inputs[self.embedding_key].astype(inputs["coordinates"].dtype)

        wi = jax.nn.softplus(
            nn.Dense(self.ncharges, use_bias=True, name="wi")(embedding)
        )

        batch_index = inputs["batch_index"]
        nsys = inputs["natoms"].shape[0]
        wtot = jax.ops.segment_sum(wi, batch_index, nsys)

        Qtot = (
            inputs[self.total_charge_key].astype(wi.dtype)
            if self.total_charge_key in inputs
            else jnp.zeros(nsys, dtype=wi.dtype)
        )
        if Qtot.ndim == 0:
            Qtot = Qtot * jnp.ones(nsys, dtype=wi.dtype)

        
        qtilde = nn.Dense(self.ncharges, use_bias=True, name="qi")(embedding)
        qtot = jax.ops.segment_sum(qtilde, batch_index, nsys)
        dq = Qtot[:, None] - qtot
        f = (dq / wtot)[batch_index]

        if "alch_group" in inputs:
            alch_group = inputs["alch_group"]
            lambda_e = inputs["alch_elambda"]
            Qligand = inputs.get("alch_ligand_charge", jnp.zeros(nsys))
            if Qligand.ndim == 0:
                Qligand = Qligand * jnp.ones(nsys, dtype=wi.dtype)

            alch_index = alch_group + 2 * batch_index
            qtot = jax.ops.segment_sum(qtilde, alch_index, 2 * nsys).reshape(
                nsys, 2, self.ncharges
            )
            wtot = jax.ops.segment_sum(wi, alch_index, 2 * nsys).reshape(
                nsys, 2, self.ncharges
            )
            # compute ligand alone
            dq = Qligand[:, None] - qtot[:, 1, :]
            f1 = (dq / wtot[:, 1, :])[batch_index]
            # q1 = alch_group*(qtilde + wi * f1[batch_index])

            # qtot = jax.ops.segment_sum(qtilde*(1-alch_group), batch_index, nsys)
            # wtot = jax.ops.segment_sum(wi*(1-alch_group), batch_index, nsys)
            # compute solvent alone
            dq = (Qtot - Qligand)[:, None] - qtot[:, 0, :]
            f0 = (dq / wtot[:, 0, :])[batch_index]
            # q0 = (1-alch_group)*(qtilde + wi * f0[batch_index])

            # alch_q = alch_group*q1 + (1-alch_group)*q0
            # q = lambda_e*q + (1-lambda_e)*alch_q

            alch_f = (1 - alch_group[:, None]) * f0 + alch_group[:, None] * f1
            f = lambda_e * f + (1 - lambda_e) * alch_f

        q = qtilde + wi * f

        if self.squeeze and self.ncharges == 1:
            q = jnp.squeeze(q, axis=-1)

        output_key = self.output_key if self.output_key is not None else self.name
        return {
            **inputs,
            output_key: q,
        }